#MIT License

#Copyright (c) 2018 XuanyiLi

#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in all
#copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.

# Based on https://github.com/meteorshowers/hed-pytorch/

import torch
import warnings
import numpy as np
import torch.nn.functional as F
import torch.nn as nn
import torch
from torchvision import models
import math
from fastai.callbacks.hooks import hook_outputs
from fastai.vision.learner import create_body
from fastai.callbacks.hooks import model_sizes
from ._arcgis_model import _get_backbone_meta
from fastprogress.fastprogress import progress_bar
from skimage.morphology import skeletonize, binary_dilation

class _HEDModel(nn.Module):
    def __init__(self, backbone_fn, chip_size=224):
        super().__init__()

        backbone_name = backbone_fn.__name__
        if "vgg" in backbone_name:
            self.backbone = create_body(backbone_fn, pretrained=True)[0]#[:-1]
        else:
            self.backbone = create_body(backbone_fn, pretrained=True)
        
        self.hookable_modules = list(self.backbone.children())

        for i, module in enumerate(self.hookable_modules):
            if isinstance(module, nn.MaxPool2d):
                module.ceil_mode = True
                module.kernel_size = 2
            elif isinstance(module, nn.Conv2d) and i==0:
                module.stride = (1, 1)

        if "vgg" in backbone_name:
            hooks = [self.hookable_modules[i-1] for i, module in enumerate(self.hookable_modules) if isinstance(module, nn.MaxPool2d)]

        else:
            hooks = [self.hookable_modules[i] for i in range(2,8) if i != 3]
        
        self.hook = hook_outputs(hooks)
        model_sizes(self.backbone, size=(chip_size, chip_size))
        layer_num_channels = [k.stored.shape[1] for k in self.hook]

        self.score_dsn1 = nn.Conv2d(layer_num_channels[0], 1, 1)
        self.score_dsn2 = nn.Conv2d(layer_num_channels[1], 1, 1)
        self.score_dsn3 = nn.Conv2d(layer_num_channels[2], 1, 1)
        self.score_dsn4 = nn.Conv2d(layer_num_channels[3], 1, 1)
        self.score_dsn5 = nn.Conv2d(layer_num_channels[4], 1, 1)
        self.score_final = nn.Conv2d(5, 1, 1)
        
    def forward(self, x):
        img_H, img_W = x.shape[2], x.shape[3]
        x = self.backbone(x)
        features = self.hook.stored
        
        so1 = self.score_dsn1(features[0])
        so2 = self.score_dsn2(features[1])
        so3 = self.score_dsn3(features[2])
        so4 = self.score_dsn4(features[3])
        so5 = self.score_dsn5(features[4])
        
        weight_deconv2 =  make_bilinear_weights(4, 1).to(x.device)
        weight_deconv3 =  make_bilinear_weights(8, 1).to(x.device)
        weight_deconv4 =  make_bilinear_weights(16, 1).to(x.device)
        weight_deconv5 =  make_bilinear_weights(32, 1).to(x.device)
        
        upsample2 = torch.nn.functional.conv_transpose2d(so2, weight_deconv2, stride=2)
        upsample3 = torch.nn.functional.conv_transpose2d(so3, weight_deconv3, stride=4)
        upsample4 = torch.nn.functional.conv_transpose2d(so4, weight_deconv4, stride=8)
        upsample5 = torch.nn.functional.conv_transpose2d(so5, weight_deconv5, stride=16)
        
        so2 = crop(upsample2, img_H, img_W)
        so3 = crop(upsample3, img_H, img_W)
        so4 = crop(upsample4, img_H, img_W)
        so5 = crop(upsample5, img_H, img_W)

        fusecat = torch.cat((so1, so2, so3, so4, so5), dim=1)
        fuse = self.score_final(fusecat)
        
        results = [so1, so2, so3, so4, so5, fuse]
        results = [torch.sigmoid(r) for r in results]
        
        return results
    
def crop(variable, th, tw):
        h, w = variable.shape[2], variable.shape[3]
        x1 = int(round((w - tw) / 2.))
        y1 = int(round((h - th) / 2.))
        return variable[:, :, y1 : y1 + th, x1 : x1 + tw]
    
def make_bilinear_weights(size, num_channels):
    factor = (size + 1) // 2
    if size % 2 == 1:
        center = factor - 1
    else:
        center = factor - 0.5
    og = np.ogrid[:size, :size]
    filt = (1 - abs(og[0] - center) / factor) * (1 - abs(og[1] - center) / factor)
    # print(filt)
    filt = torch.from_numpy(filt)
    w = torch.zeros(num_channels, num_channels, size, size)
    w.requires_grad = False
    for i in range(num_channels):
        for j in range(num_channels):
            if i == j:
                w[i, j] = filt
    return w

def cross_entropy_loss(prediction, label):
    
    label = label.long()
    mask = (label != 0).float()
    num_positive = torch.sum(mask).float()
    num_negative = mask.numel() - num_positive
    mask[mask != 0] = num_negative / (num_positive + num_negative)
    mask[mask == 0] = num_positive / (num_positive + num_negative)
    cost = torch.nn.functional.binary_cross_entropy(
            prediction.float(),label.float(), weight=mask)#, reduce=False)
    return torch.sum(cost)

def hed_loss(out, labels):
    loss = 0.
    for o in out:
        loss = loss + cross_entropy_loss(o, labels)
    return loss

def accuracy(input, target):
    if isinstance(input, tuple): # while training
        input = input[0]
    target = target.byte().squeeze(1)
    input = input[-1]
    input = (input>=0.5).byte().squeeze(1)
    return (input == target).float().mean()

def get_true_positive(mask1, mask2, buffer):
    tp = 0
    indices = np.where(mask1 == 1)
    for ind in range(len(indices[0])):
        tp += np.any(
            mask2[max(indices[0][ind]-buffer,0): indices[0][ind]+buffer+1,
              max(indices[1][ind]-buffer,0): indices[1][ind]+buffer+1]).astype(np.int)
    return tp

def get_confusion_metric(gt, pred, buffer):
    
    tp, predicted_tp, actual_tp = 0, 0, 0
    for i in range(gt.shape[0]):
        gt_mask = skeletonize(binary_dilation(gt[i]))
        pred_mask = skeletonize(binary_dilation(pred[i]))
        tp += get_true_positive(gt_mask, pred_mask, buffer)
        predicted_tp += len(np.where(pred_mask == 1)[0])
        actual_tp += len(np.where(gt_mask == 1)[0])
    
    return tp, predicted_tp, actual_tp

def f1_score(pred, gt):

    gt = gt.byte().squeeze(1).cpu().numpy()
    pred = (pred[-1]>=0.5).byte().squeeze(1).cpu().numpy()
    tp, predicted_tp, actual_tp = get_confusion_metric(gt, pred, 3)
    precision = tp/(predicted_tp + 1e-12)
    recall = tp/(actual_tp + 1e-12)
    f1score = 2*precision*recall/(precision + recall + 1e-12)
    return torch.tensor(f1score)

def accuracies(model, dl, detect_thresh=0.5, buffer=3, show_progress=True):

    precision, recall, f1score = [], [], []
    model.learn.model.eval()
    acc = {}
    with torch.no_grad():
        for input, gt in progress_bar(dl, display=show_progress):
            predictions = model.learn.model(input)
            gt = gt.byte().squeeze(1).cpu().numpy()
            pred = (predictions[-1]>=detect_thresh).byte().squeeze(1).cpu().numpy()
            tp, predicted_tp, actual_tp = get_confusion_metric(gt, pred, buffer)
            prec = tp/(predicted_tp + 1e-12)
            rec = tp/(actual_tp + 1e-12)
            precision.append(prec)
            recall.append(rec)
            f1score.append(2*prec*rec/(prec + rec + 1e-12))
    acc['Precision'] = np.mean(precision)
    acc['Recall'] = np.mean(recall)
    acc['F1 Score'] = np.mean(f1score)

    return acc



#Code from https://github.com/fastai/course-v3/blob/master/nbs/dl2/cyclegan_ws.ipynb
import torch
from fastai.vision import nn, Callable, List, LearnerCallback, optim, ifnone, F, flatten_model, requires_grad, SmoothenValue, add_metrics
from .._utils.cyclegan import calculate_activation_statistics, calculate_frechet_distance
def convT_norm_relu(ch_in:int, ch_out:int, norm_layer:nn.Module, ks:int=3, stride:int=2, bias:bool=True):
    return [nn.ConvTranspose2d(ch_in, ch_out, kernel_size=ks, stride=stride, padding=1, output_padding=1, bias=bias),
            norm_layer(ch_out), nn.ReLU(True)]

def pad_conv_norm_relu(ch_in:int, ch_out:int, pad_mode:str, norm_layer:nn.Module, ks:int=3, bias:bool=True, 
                       pad=1, stride:int=1, activ:bool=True, init:Callable=nn.init.kaiming_normal_)->List[nn.Module]:
    layers = []
    if pad_mode == 'reflection': layers.append(nn.ReflectionPad2d(pad))
    elif pad_mode == 'border':   layers.append(nn.ReplicationPad2d(pad))
    p = pad if pad_mode == 'zeros' else 0
    conv = nn.Conv2d(ch_in, ch_out, kernel_size=ks, padding=p, stride=stride, bias=bias)
    if init:
        init(conv.weight)
        if hasattr(conv, 'bias') and hasattr(conv.bias, 'data'): conv.bias.data.fill_(0.)
    layers += [conv, norm_layer(ch_out)]
    if activ: layers.append(nn.ReLU(inplace=True))
    return layers

class ResnetBlock(nn.Module):
    def __init__(self, dim:int, pad_mode:str='reflection', norm_layer:nn.Module=None, dropout:float=0., bias:bool=True):
        super().__init__()
        assert pad_mode in ['zeros', 'reflection', 'border'], f'padding {pad_mode} not implemented.'
        norm_layer = ifnone(norm_layer, nn.InstanceNorm2d)
        layers = pad_conv_norm_relu(dim, dim, pad_mode, norm_layer, bias=bias)
        if dropout != 0: layers.append(nn.Dropout(dropout))
        layers += pad_conv_norm_relu(dim, dim, pad_mode, norm_layer, bias=bias, activ=False)
        self.conv_block = nn.Sequential(*layers)

    def forward(self, x): return x + self.conv_block(x)

def resnet_generator(ch_in:int, ch_out:int, n_ftrs:int=64, norm_layer:nn.Module=None, 
                     dropout:float=0., n_blocks:int=6, pad_mode:str='reflection')->nn.Module:
    norm_layer = ifnone(norm_layer, nn.InstanceNorm2d)
    bias = (norm_layer == nn.InstanceNorm2d)
    layers = pad_conv_norm_relu(ch_in, n_ftrs, 'reflection', norm_layer, pad=3, ks=7, bias=bias)
    for i in range(2):
        layers += pad_conv_norm_relu(n_ftrs, n_ftrs *2, 'zeros', norm_layer, stride=2, bias=bias)
        n_ftrs *= 2
    layers += [ResnetBlock(n_ftrs, pad_mode, norm_layer, dropout, bias) for _ in range(n_blocks)]
    for i in range(2):
        layers += convT_norm_relu(n_ftrs, n_ftrs//2, norm_layer, bias=bias)
        n_ftrs //= 2
    layers += [nn.ReflectionPad2d(3), nn.Conv2d(n_ftrs, ch_out, kernel_size=7, padding=0), nn.Tanh()]
    return nn.Sequential(*layers)

def conv_norm_lr(ch_in:int, ch_out:int, norm_layer:nn.Module=None, ks:int=3, bias:bool=True, pad:int=1, stride:int=1, 
                 activ:bool=True, slope:float=0.2, init:Callable=nn.init.kaiming_normal_)->List[nn.Module]:
    conv = nn.Conv2d(ch_in, ch_out, kernel_size=ks, padding=pad, stride=stride, bias=bias)
    if init:
        init(conv.weight)
        if hasattr(conv, 'bias') and hasattr(conv.bias, 'data'): conv.bias.data.fill_(0.)
    layers = [conv]
    if norm_layer is not None: layers.append(norm_layer(ch_out))
    if activ: layers.append(nn.LeakyReLU(slope, inplace=True))
    return layers

def discriminator(ch_in:int, n_ftrs:int=64, n_layers:int=3, norm_layer:nn.Module=None, sigmoid:bool=False)->nn.Module:
    norm_layer = ifnone(norm_layer, nn.InstanceNorm2d)
    bias = (norm_layer == nn.InstanceNorm2d)
    layers = conv_norm_lr(ch_in, n_ftrs, ks=4, stride=2, pad=1)
    for i in range(n_layers-1):
        new_ftrs = 2*n_ftrs if i <= 3 else n_ftrs
        layers += conv_norm_lr(n_ftrs, new_ftrs, norm_layer, ks=4, stride=2, pad=1, bias=bias)
        n_ftrs = new_ftrs
    new_ftrs = 2*n_ftrs if n_layers <=3 else n_ftrs
    layers += conv_norm_lr(n_ftrs, new_ftrs, norm_layer, ks=4, stride=1, pad=1, bias=bias)
    layers.append(nn.Conv2d(new_ftrs, 1, kernel_size=4, stride=1, padding=1))
    if sigmoid: layers.append(nn.Sigmoid())
    return nn.Sequential(*layers)

class CycleGAN(nn.Module):
    
    def __init__(self, ch_in:int, ch_out:int, n_features:int=64, disc_layers:int=3, gen_blocks:int=6, lsgan:bool=True, 
                 drop:float=0., norm_layer:nn.Module=None):
        super().__init__()
        self.D_A = discriminator(ch_in, n_features, disc_layers, norm_layer, sigmoid=not lsgan)
        self.D_B = discriminator(ch_in, n_features, disc_layers, norm_layer, sigmoid=not lsgan)
        self.G_A = resnet_generator(ch_in, ch_out, n_features, norm_layer, drop, gen_blocks)
        self.G_B = resnet_generator(ch_in, ch_out, n_features, norm_layer, drop, gen_blocks)
        #G_A: takes real input B and generates fake input A
        #G_B: takes real input A and generates fake input B
        #D_A: trained to make the difference between real input A and fake input A
        #D_B: trained to make the difference between real input B and fake input B
        self.arcgis_results = False
    
    def forward(self, real_A, real_B):
        fake_A, fake_B = self.G_A(real_B), self.G_B(real_A)
        if self.arcgis_results: return torch.cat([fake_A[:,None],fake_B[:,None]], 1)
        idt_A, idt_B = self.G_A(real_A), self.G_B(real_B) #Needed for the identity loss during training.
        return [fake_A, fake_B, idt_A, idt_B]

class AdaptiveLoss(nn.Module):
    def __init__(self, crit):
        super().__init__()
        self.crit = crit
    
    def forward(self, output, target:bool, **kwargs):
        targ = output.new_ones(*output.size()) if target else output.new_zeros(*output.size())
        return self.crit(output, targ, **kwargs)

class CycleGanLoss(nn.Module):
    
    def __init__(self, cgan:nn.Module, lambda_A:float=10., lambda_B:float=10, lambda_idt:float=0.5, lsgan:bool=True):
        super().__init__()
        self.cgan,self.l_A,self.l_B,self.l_idt = cgan,lambda_A,lambda_B,lambda_idt
        self.crit = AdaptiveLoss(F.mse_loss if lsgan else F.binary_cross_entropy)
    
    def set_input(self, input):
        self.real_A,self.real_B = input
    
    def forward(self, output, target):
        fake_A, fake_B, idt_A, idt_B = output
        #Generators should return identity on the datasets they try to convert to
        self.id_loss = self.l_idt * (self.l_A * F.l1_loss(idt_A, self.real_A) + self.l_B * F.l1_loss(idt_B, self.real_B))
        #Generators are trained to trick the discriminators so the following should be ones
        self.gen_loss = self.crit(self.cgan.D_A(fake_A), True) + self.crit(self.cgan.D_B(fake_B), True)
        #Cycle loss
        self.cyc_loss  = self.l_A * F.l1_loss(self.cgan.G_A(fake_B), self.real_A)
        self.cyc_loss += self.l_B * F.l1_loss(self.cgan.G_B(fake_A), self.real_B)
        return self.id_loss+self.gen_loss+self.cyc_loss

class CycleGANTrainer(LearnerCallback):
    _order = -20 #Need to run before the Recorder
    
    def _set_trainable(self, D_A=False, D_B=False):
        gen = (not D_A) and (not D_B)
        requires_grad(self.learn.model.G_A, gen)
        requires_grad(self.learn.model.G_B, gen)
        requires_grad(self.learn.model.D_A, D_A)
        requires_grad(self.learn.model.D_B, D_B)
        if not gen:
            self.opt_D_A.lr, self.opt_D_A.mom = self.learn.opt.lr, self.learn.opt.mom
            self.opt_D_A.wd, self.opt_D_A.beta = self.learn.opt.wd, self.learn.opt.beta
            self.opt_D_B.lr, self.opt_D_B.mom = self.learn.opt.lr, self.learn.opt.mom
            self.opt_D_B.wd, self.opt_D_B.beta = self.learn.opt.wd, self.learn.opt.beta
    
    def on_train_begin(self, **kwargs):
        self.G_A,self.G_B = self.learn.model.G_A,self.learn.model.G_B
        self.D_A,self.D_B = self.learn.model.D_A,self.learn.model.D_B
        self.crit = self.learn.loss_func.crit
        if not getattr(self,'opt_G',None):
            self.opt_G = self.learn.opt.new([nn.Sequential(*flatten_model(self.G_A), *flatten_model(self.G_B))])
        else: 
            self.opt_G.lr,self.opt_G.wd = self.opt.lr,self.opt.wd
            self.opt_G.mom,self.opt_G.beta = self.opt.mom,self.opt.beta
        if not getattr(self,'opt_D_A',None):
            self.opt_D_A = self.learn.opt.new([nn.Sequential(*flatten_model(self.D_A))])
        if not getattr(self,'opt_D_B',None):
            self.opt_D_B = self.learn.opt.new([nn.Sequential(*flatten_model(self.D_B))])
        self.learn.opt.opt = self.opt_G.opt
        self._set_trainable()
        self.id_smter,self.gen_smter,self.cyc_smter = SmoothenValue(0.98),SmoothenValue(0.98),SmoothenValue(0.98)
        self.da_smter,self.db_smter = SmoothenValue(0.98),SmoothenValue(0.98)
        self.recorder.add_metric_names(['id_loss', 'gen_loss', 'cyc_loss', 'D_A_loss', 'D_B_loss'])
        
    def on_batch_begin(self, last_input, **kwargs):
        self.learn.loss_func.set_input(last_input)
    
    def on_backward_begin(self, **kwargs):
        self.id_smter.add_value(self.loss_func.id_loss.detach().cpu())
        self.gen_smter.add_value(self.loss_func.gen_loss.detach().cpu())
        self.cyc_smter.add_value(self.loss_func.cyc_loss.detach().cpu())
    
    def on_batch_end(self, last_input, last_output, **kwargs):
        self.G_A.zero_grad(); self.G_B.zero_grad()
        fake_A, fake_B = last_output[0].detach(), last_output[1].detach()
        real_A, real_B = last_input
        self._set_trainable(D_A=True)
        self.D_A.zero_grad()
        loss_D_A = 0.5 * (self.crit(self.D_A(real_A), True) + self.crit(self.D_A(fake_A), False))
        self.da_smter.add_value(loss_D_A.detach().cpu())
        if self.learn.model.training == True:
            loss_D_A.backward()
        #loss_D_A.backward()
        self.opt_D_A.step()
        self._set_trainable(D_B=True)
        self.D_B.zero_grad()
        loss_D_B = 0.5 * (self.crit(self.D_B(real_B), True) + self.crit(self.D_B(fake_B), False))
        self.db_smter.add_value(loss_D_B.detach().cpu())
        if self.learn.model.training == True:
            loss_D_B.backward()
        #loss_D_B.backward()
        self.opt_D_B.step()
        self._set_trainable()
        
    def on_epoch_end(self, last_metrics, **kwargs):
        return add_metrics(last_metrics, [s.smooth for s in [self.id_smter,self.gen_smter,self.cyc_smter,
                                                             self.da_smter,self.db_smter]])

def compute_fid_metric(model, data):
    input_a = []
    input_b = []
    pred_a = []
    pred_b = []
    for input, target in data.valid_dl:
        input_a.append(input[0]/2+0.5)
        input_b.append(input[1]/2+0.5)
        pred = model.learn.pred_batch(batch=(input,target))
        pred_a.append(pred[0]/2+0.5)
        pred_b.append(pred[1]/2+0.5)
    
    data_len = len(data.valid_ds)
    batch_size = data.batch_size

    m1_a, s1_a = calculate_activation_statistics(batch_size, data_len, input_a)
    m2_a, s2_a = calculate_activation_statistics(batch_size, data_len, pred_a)

    m1_b, s1_b = calculate_activation_statistics(batch_size, data_len, input_b)
    m2_b, s2_b = calculate_activation_statistics(batch_size, data_len, pred_b)

    fid_value_a = calculate_frechet_distance(m1_a, s1_a, m2_a, s2_a)
    fid_value_b = calculate_frechet_distance(m1_b, s1_b, m2_b, s2_b)

    return fid_value_a, fid_value_b


import os
import tempfile
import traceback
import json
import warnings
import pickle
import math
from pathlib import Path

HAS_FASTAI = True
import_exception=None

import arcgis
from arcgis.features import FeatureLayer

try:
    from ._arcgis_model import ArcGISModel, _raise_fastai_import_error
    from fastai.basic_train import Learner, load_learner
    from fastprogress.fastprogress import progress_bar
    from .._utils.tabular_data import TabularDataObject
    from fastai.torch_core import split_model_idx
    import torch
    from fastai.metrics import r2_score
    from ._tsmodel_archs._InceptionTime import _TSInceptionTime
    from ._tsmodel_archs._Resnet import _TSResNet
    from ._tsmodel_archs._ResCNN import _TSResCNN
    from ._tsmodel_archs._FCN import _TSFCN
    from ._tsmodel_archs._LSTM import _TSLSTM
    from .._utils.TSData import To3dTensor, ToTensor
    _model_arch = {
        'inceptiontime': _TSInceptionTime,
        'resnet': _TSResNet,
        'rescnn': _TSResCNN,
        'fcn': _TSFCN,
        'lstm': _TSLSTM
    }
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_FASTAI = False
    _model_arch = {}

HAS_NUMPY = True
try:
    import numpy as np
except:
    HAS_NUMPY = False

_PROTOCOL_LEVEL = 2

HAS_PANDAS = True
try:
    import pandas as pd
except:
    HAS_PANDAS = False


def _get_model_from_path(pretrained_path):
    learn = load_learner(
        os.path.dirname(pretrained_path),
        os.path.basename(pretrained_path).split('.')[0] + "_exported.pth",
        no_check=True
    )

    return learn


class TimeSeriesModel(ArcGISModel):
    """
    Creates a TimeSeriesModel Object.
    Based on the Fast.ai's https://github.com/timeseriesAI/timeseriesAI

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required TabularDataObject. Returned data object from
                            `prepare_tabulardata` function.
    ---------------------   -------------------------------------------
    seq_len                 Required Integer. Sequence Length for the series.
                            In case of raster only, seq_len = number of rasters,
                            any other passed value will be ignored.
    ---------------------   -------------------------------------------
    model_arch              Optional string. Model Architecture.
                            Allowed "InceptionTime", "ResCNN",
                            "Resnet", "FCN"
    ---------------------   -------------------------------------------
    **kwargs                Optional kwargs.
    =====================   ===========================================

    :returns: `TimeSeriesModel` Object
    """

    def __init__(self, data, seq_len, model_arch='InceptionTime', **kwargs):

        data_bunch = None

        if not data._is_empty:
            data_bunch = data._time_series_bunch(seq_len)

        super().__init__(data, None)

        if not data_bunch:
            self.learn = _get_model_from_path(kwargs.get('pretrained_path'))
        elif kwargs.get('pretrained_path'):
            self.learn = _get_model_from_path(kwargs.get('pretrained_path'))
            self.learn.data = data_bunch
        else:
            if not _model_arch.get(model_arch.lower()):
                raise Exception("Invalid model architecture")

            model_arch_ob = _model_arch.get(model_arch.lower())
            if model_arch.lower() == 'lstm':
                model = model_arch_ob(data_bunch.features, data_bunch.c, self._device, **kwargs).to(self._device)
            else:
                if model_arch.lower() in ['resnet', 'fcn']:
                    kwargs['device'] = self._device
                model = model_arch_ob(data_bunch.features, data_bunch.c, **kwargs).to(self._device)
                if model_arch.lower() in ['resnet', 'fcn']:
                    del kwargs['device']
            self.learn = Learner(data_bunch, model, model_dir=tempfile.TemporaryDirectory().name)
            self.learn.data = data_bunch

        self.learn.layer_groups = split_model_idx(self.learn.model, [1])
        self._model_arch = model_arch.lower()
        if kwargs.get('pretrained_path'):
            del kwargs['pretrained_path']
        self._kwargs = kwargs
        self._seq_len = seq_len

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a TimeSeriesModel Object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_tabulardata` function or None for
                                inferencing.
        =====================   ===========================================

        :returns: `TimeSeriesModel` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = Path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        dependent_variable = emd['dependent_variable']
        categorical_variables = emd['categorical_variables']
        continuous_variables = emd['continuous_variables']

        _is_classification = False
        if emd['_is_classification'] == "classification":
            _is_classification = True

        model_params = emd['model_params']
        model_arch = emd['model_arch']
        seq_len = emd['seq_len']
        index_field = emd.get('index_field', None)
        # encoder_path = os.path.join(os.path.dirname(emd_path),
        #                             os.path.basename(emd_path).split('.')[0] + '_encoders.pkl')

        scaler_path = os.path.join(os.path.dirname(emd_path),
                                    os.path.basename(emd_path).split('.')[0] + '_scaler.pkl')

        encoder_mapping = None
        scaler = None

        # if os.path.exists(encoder_path):
        #     with open(encoder_path, 'rb') as f:
        #         encoder_mapping = pickle.loads(f.read())

        if os.path.exists(scaler_path):
            with open(scaler_path, 'rb') as f:
                scaler = pickle.loads(f.read())

        if data is None:
            data = TabularDataObject._empty(categorical_variables, continuous_variables, dependent_variable, encoder_mapping)
            data._is_classification = _is_classification
            data._column_transforms_mapping = scaler

            if index_field is not None:
                data._index_field = index_field

            class_object = cls(data, seq_len, model_arch=model_arch, pretrained_path=emd_path, **model_params)
            class_object._data.emd = emd
            class_object._data.emd_path = emd_path
            return class_object

        return cls(data, seq_len, model_arch=model_arch, pretrained_path=emd_path, **model_params)

    def save(self, name_or_path, framework='PyTorch', publish=False, gis=None, save_optimizer=False, **kwargs):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment to Image Server or ArcGIS Pro.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Folder path to save the model.
        ---------------------   -------------------------------------------
        framework               Optional string. Defines the framework of the
                                model. (Only supported by ``SingleShotDetector``, currently.)
                                If framework used is ``TF-ONNX``, ``batch_size`` can be
                                passed as an optional keyword argument.

                                Framework choice: 'PyTorch' and 'TF-ONNX'
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
        ---------------------   -------------------------------------------
        gis                     Optional GIS Object. Used for publishing the item.
                                If not specified then active gis user is taken.
        ---------------------   -------------------------------------------
        save_optimizer          Optional boolean. Used for saving the model-optimizer
                                state along with the model. Default is set to False
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise, default False.
        =====================   ===========================================
        """

        if '\\' in name_or_path or '/' in name_or_path:
            path = name_or_path
        else:
            path = os.path.join(os.getcwd(), name_or_path)

        if not os.path.exists(path):
            os.mkdir(path)

        base_file_name = os.path.basename(path)

        TimeSeriesModel._save_encoders(self._data._column_transforms_mapping, path, base_file_name)

        self.learn.export(os.path.join(path, os.path.basename(path) + '_exported.pth'))
        from IPython.utils import io
        with io.capture_output() as captured:
            super().save(path, framework, publish, gis, save_optimizer=save_optimizer, **kwargs)

        return Path(path)

    @staticmethod
    def _save_encoders(scaler, path, base_file_name):
        # if not encoder_mapping:
        #     return

        # encoder_file = os.path.join(path, base_file_name + '_encoders.pkl')
        # with open(encoder_file, 'wb') as f:
        #     f.write(pickle.dumps(encoder_mapping, protocol=_PROTOCOL_LEVEL))

        scaler_file = os.path.join(path, base_file_name + '_scaler.pkl')
        with open(scaler_file, 'wb') as f:
            f.write(pickle.dumps(scaler, protocol=_PROTOCOL_LEVEL))

    @property
    def _model_metrics(self):
        # from IPython.utils import io
        # with io.capture_output() as captured:
        #     score = self.score()

        return {'score': self.score()}

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        _emd_template["ModelType"] = "TimeSeriesModel"
        _emd_template["model_arch"] = self._model_arch
        _emd_template['model_params'] = self._kwargs
        _emd_template['seq_len'] = self._seq_len
        _emd_template["dependent_variable"] = self._data._dependent_variable
        _emd_template["categorical_variables"] = self._data._categorical_variables
        _emd_template["continuous_variables"] = self._data._continuous_variables

        if self._data._index_field:
            _emd_template['index_field'] = self._data._index_field

        _emd_template['_is_classification'] = "classification" if self._data._is_classification else "regression"

        return _emd_template

    def _predict(self, orig_sequence):
        sequence = np.array(orig_sequence, dtype='float32')
        if sequence.shape[0] == 1:
            seq_arr = To3dTensor(orig_sequence).to(self._device)
        else:
            seq_arr = ToTensor(np.expand_dims(np.array(sequence, dtype='float32'), axis=0)).to(self._device)
        model = self.learn.model
        model.eval()

        with torch.no_grad():
            output = model(seq_arr).item()

        return output

    def predict(
            self,
            input_features=None,
            explanatory_rasters=None,
            datefield=None,
            distance_features=None,
            output_layer_name="Prediction Layer",
            gis=None,
            prediction_type='features',
            output_raster_path=None,
            match_field_names=None,
            number_of_predictions=None
    ):
        """

        Predict on data from feature layer and or raster data.

        =================================   =========================================================================
        **Argument**                        **Description**
        ---------------------------------   -------------------------------------------------------------------------
        input_features                      Optional Feature Layer or spatially enabled dataframe.
                                            Contains features with location of the input data.
                                            Required if prediction_type is 'features' or 'dataframe'
        ---------------------------------   -------------------------------------------------------------------------
        explanatory_rasters                 Optional list of Raster Objects.
                                            Required if prediction_type is 'rasters'
        ---------------------------------   -------------------------------------------------------------------------
        datefield                           Optional field_name.
                                            This field contains the date in the input_features.
                                            The field type can be a string or date time field.
                                            If specified, the field will be split into
                                            Year, month, week, day, dayofweek, dayofyear,
                                            is_month_end, is_month_start, is_quarter_end,
                                            is_quarter_start, is_year_end, is_year_start,
                                            hour, minute, second, elapsed and these will be added
                                            to the prepared data as columns.
                                            All fields other than elapsed and dayofyear are treated
                                            as categorical.
        ---------------------------------   -------------------------------------------------------------------------
        distance_features                   Optional List of Feature Layer objects.
                                            These layers are used for calculation of field "NEAR_DIST_1",
                                            "NEAR_DIST_2" etc in the output dataframe.
                                            These fields contain the nearest feature distance
                                            from the input_features.
                                            Same as `prepare_tabulardata()`.
        ---------------------------------   -------------------------------------------------------------------------
        output_layer_name                   Optional string. Used for publishing the output layer.
        ---------------------------------   -------------------------------------------------------------------------
        gis                                 Optional GIS Object. Used for publishing the item.
                                            If not specified then active gis user is taken.
        ---------------------------------   -------------------------------------------------------------------------
        prediction_type                     Optional String.
                                            Set 'features' or 'dataframe' to make output predictions.
        ---------------------------------   -------------------------------------------------------------------------
        output_raster_path                  Optional path. Required when prediction_type='raster', saves
                                            the output raster to this path.
        ---------------------------------   -------------------------------------------------------------------------
        match_field_names                   Optional string.
                                            Specify mapping of the original training set with prediction set.
        ---------------------------------   -------------------------------------------------------------------------
        number_of_predictions               Optional int for univariate time series.
                                            Specify the number of predictions to make, adds new rows to the dataframe.
                                            For multivariate or if None, it expects the dataframe to have empty rows.
                                            For prediction_type='raster', a new raster is created.
        =================================   =========================================================================

        :returns Feature Layer/dataframe if prediction_type='features'/'dataframe', else returns True and saves output
        raster at the specified path.
        """

        rasters = explanatory_rasters if explanatory_rasters else []
        if prediction_type in ['features', 'dataframe']:

            if input_features is None:
                raise Exception("Feature Layer required for predict_features=True")

            gis = gis if gis else arcgis.env.active_gis
            return self._predict_features(input_features, rasters, datefield, distance_features, output_layer_name, gis,
                                          match_field_names, number_of_predictions, prediction_type)
        else:
            if not rasters:
                raise Exception("Rasters required for predict_features=False")

            if not output_raster_path:
                raise Exception("Please specify output_raster_folder_path to save the output.")

            return self._predict_rasters(output_raster_path, rasters, match_field_names)

    def _predict_rasters(self, output_raster_path, rasters, match_field_names=None):
        if not os.path.exists(os.path.dirname(output_raster_path)):
            raise Exception("Output directory doesn't exist")

        if os.path.exists(output_raster_path):
            raise Exception("Output Folder already exists")

        try:
            import arcpy
        except:
            raise Exception("This function requires arcpy.")

        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")

        try:
            import pandas as pd
        except:
            raise Exception("This function requires pandas.")

        fields_needed = self._data._categorical_variables + self._data._continuous_variables

        try:
            arcpy.env.outputCoordinateSystem = rasters[0].extent['spatialReference']['wkt']
        except:
            arcpy.env.outputCoordinateSystem = rasters[0].extent['spatialReference']['wkid']

        xmin = rasters[0].extent['xmin']
        xmax = rasters[0].extent['xmax']
        ymin = rasters[0].extent['ymin']
        ymax = rasters[0].extent['ymax']
        min_cell_size_x = rasters[0].mean_cell_width
        min_cell_size_y = rasters[0].mean_cell_height

        default_sr = rasters[0].extent['spatialReference']

        for raster in rasters:
            point_upper = arcgis.geometry.Point(
                {'x': raster.extent['xmin'], 'y': raster.extent['ymax'], 'sr': raster.extent['spatialReference']})
            point_lower = arcgis.geometry.Point(
                {'x': raster.extent['xmax'], 'y': raster.extent['ymin'], 'sr': raster.extent['spatialReference']})
            cell_size = arcgis.geometry.Point(
                {'x': raster.mean_cell_width, 'y': raster.mean_cell_height, 'sr': raster.extent['spatialReference']})

            points = arcgis.geometry.project([point_upper, point_lower, cell_size], raster.extent['spatialReference'],
                                             default_sr)
            point_upper = points[0]
            point_lower = points[1]
            cell_size = points[2]

            if xmin > point_upper.x:
                xmin = point_upper.x
            if ymax < point_upper.y:
                ymax = point_upper.y
            if xmax < point_lower.x:
                xmax = point_lower.x
            if ymin > point_lower.y:
                ymin = point_lower.y

            if min_cell_size_x > cell_size.x:
                min_cell_size_x = cell_size.x

            if min_cell_size_y > cell_size.y:
                min_cell_size_y = cell_size.y

        max_raster_columns = int(abs(math.ceil((xmax - xmin) / min_cell_size_x)))
        max_raster_rows = int(abs(math.ceil((ymax - ymin) / min_cell_size_y)))

        point_upper = arcgis.geometry.Point({'x': xmin, 'y': ymax, 'sr': default_sr})
        cell_size = arcgis.geometry.Point({'x': min_cell_size_x, 'y': min_cell_size_y, 'sr': default_sr})

        raster_data = {}
        for raster in rasters:
            field_name = raster.name
            point_upper_translated = \
            arcgis.geometry.project([point_upper], default_sr, raster.extent['spatialReference'])[0]
            cell_size_translated = arcgis.geometry.project([cell_size], default_sr, raster.extent['spatialReference'])[
                0]
            if field_name in fields_needed:
                raster_read = raster.read(
                    origin_coordinate=(point_upper_translated.x, point_upper_translated.y), ncols=max_raster_columns,
                    nrows=max_raster_rows, cell_size=(cell_size_translated.x, cell_size_translated.y))
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f'_{index}'
                            if not raster_data.get(key):
                                raster_data[key] = []
                            index = index + 1
                            raster_data[key].append(value)
            elif match_field_names and match_field_names.get(raster.name):
                field_name = match_field_names.get(raster.name)
                raster_read = raster.read(
                    origin_coordinate=(point_upper_translated.x, point_upper_translated.y), ncols=max_raster_columns,
                    nrows=max_raster_rows, cell_size=(cell_size_translated.x, cell_size_translated.y))
                for row in range(max_raster_rows):
                    for column in range(max_raster_columns):
                        values = raster_read[row][column]
                        index = 0
                        for value in values:
                            key = field_name
                            if index != 0:
                                key = key + f'_{index}'
                            if not raster_data.get(key):
                                raster_data[key] = []

                            index = index + 1
                            raster_data[key].append(value)
            else:
                continue

        for field in fields_needed:
            if field not in list(raster_data.keys()) and match_field_names and match_field_names.get(field, None) is None:
                raise Exception(f"Field missing {field}")

        length_values = len(raster_data[list(raster_data.keys())[0]])
        processed_output = []
        for i in progress_bar(range(length_values)):
            processed_row = []
            for raster_name in sorted(raster_data.keys()):
                processed_row.append(raster_data[raster_name][i])
            processed_output.append(self._predict([processed_row]))

        processed_numpy = np.array(processed_output, dtype='float64')
        processed_numpy = processed_numpy.reshape([max_raster_rows, max_raster_columns])
        processed_raster = arcpy.NumPyArrayToRaster(processed_numpy, arcpy.Point(xmin, ymin),
                                                    x_cell_size=min_cell_size_x, y_cell_size=min_cell_size_y)
        processed_raster.save(output_raster_path)

        return True

    def _predict_features(self, input_features, rasters=None, datefield=None, distance_features=None, output_layer_name='Prediction Layer', gis=None, match_field_names=None, number_of_predictions=None, prediction_type='features'):
        if not HAS_PANDAS:
            raise Exception("This function requires pandas library")

        if isinstance(input_features, FeatureLayer):
            orig_dataframe = input_features.query().sdf
        else:
            orig_dataframe = input_features.copy()

        if match_field_names is None:
            match_field_names = {}

        from pandas.api.types import is_datetime64_any_dtype as is_datetime

        if number_of_predictions is not None and number_of_predictions > 0:
            delta = None
            index_field_name = None
            end_value = None
            if self._data._index_field is not None:
                index_field_name = match_field_names.get(self._data._index_field, self._data._index_field)
                if index_field_name in list(orig_dataframe.columns) and is_datetime(orig_dataframe[index_field_name]):
                    delta = orig_dataframe[index_field_name].iloc[1] - orig_dataframe[index_field_name].iloc[0]
                    end_value = None
                    if delta is not None:
                        end_value = orig_dataframe[index_field_name].iloc[len(orig_dataframe)-1]
            for i in range(number_of_predictions):
                orig_dataframe = orig_dataframe.append(pd.Series(), ignore_index=True)
                if delta is not None:
                    orig_dataframe[index_field_name].iloc[len(orig_dataframe)-1] = end_value + delta.to_timedelta64()
                    end_value = end_value + delta.to_timedelta64()

        if match_field_names and match_field_names.get(self._data._dependent_variable):
            prediction_sequence_orig = orig_dataframe[match_field_names.get(self._data._dependent_variable)]
        else:
            prediction_sequence_orig = orig_dataframe[self._data._dependent_variable]

        dataframe = orig_dataframe.copy()

        fields_needed = self._data._categorical_variables + self._data._continuous_variables + [self._data._dependent_variable]
        distance_feature_layers = distance_features if distance_features else []

        continuous_variables = self._data._continuous_variables

        feature_layer_columns = []
        for column in dataframe.columns:
            column_name = column
            categorical = False

            if column_name in fields_needed:
                if column_name not in continuous_variables:
                    categorical = True
            elif match_field_names and match_field_names.get(column_name):
                if match_field_names.get(column_name) not in continuous_variables:
                    categorical = True
            else:
                continue

            feature_layer_columns.append((column_name, categorical))

        raster_columns = []
        if rasters:
            for raster in rasters:
                column_name = raster.name
                categorical = False
                if column_name in fields_needed:
                    if column_name not in continuous_variables:
                        categorical = True
                elif match_field_names and match_field_names.get(column_name):
                    column_name = match_field_names.get(column_name)
                    if column_name not in continuous_variables:
                        categorical = True
                else:
                    continue

                raster_columns.append((raster, categorical))
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            processed_dataframe, fields_mapping = TabularDataObject._prepare_dataframe_from_features(
                orig_dataframe,
                self._data._dependent_variable,
                feature_layer_columns,
                raster_columns,
                datefield,
                distance_feature_layers
            )

        if match_field_names:
            processed_dataframe.rename(columns=match_field_names, inplace=True)

        for field in fields_needed:
            if field not in processed_dataframe.columns:
                raise Exception(f"Field missing {field}")

        for column in processed_dataframe.columns:
            if column not in fields_needed:
                processed_dataframe = processed_dataframe.drop(column, axis=1)

        processed_dataframe = processed_dataframe.reindex(sorted(processed_dataframe.columns), axis=1)

        index = self._seq_len
        processed_dataframe[self._data._dependent_variable] = processed_dataframe[self._data._dependent_variable].replace(r'^\s*$', np.nan, regex=True)

        processed_dataframe_transform = processed_dataframe.copy()

        for col in list(processed_dataframe.columns):
            transformed_data = processed_dataframe[col]
            for transform in self._data._column_transforms_mapping.get(col, []):
                transformed_data = transform.fit_transform(
                    np.array(transformed_data, dtype=processed_dataframe[col].dtype).reshape(-1, 1))
                transformed_data = transformed_data.squeeze(1)
            processed_dataframe_transform[col] = np.array(transformed_data, dtype=processed_dataframe[col].dtype)

        big_bunch = []
        prediction_sequence_list = None
        processed_dataframe_transform = processed_dataframe_transform.values

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            for col in range(len(processed_dataframe.columns.values)):
                if list(processed_dataframe.columns.values)[col] == self._data._dependent_variable:
                    # big_bunch.append(prediction_sequence_list[index - self._seq_len:(index - self._seq_len + self._seq_len)])
                    prediction_sequence_list = processed_dataframe_transform[:, col]
                    big_bunch.append(prediction_sequence_list[0:self._seq_len])
                else:
                    big_bunch.append(processed_dataframe_transform[:, col][0:self._seq_len])

        if len(prediction_sequence_list) < self._seq_len:
            raise Exception("Basic Sequence not found!")

        while index < len(prediction_sequence_list):
            if prediction_sequence_list[index] in ["", None, "null", "None"] or np.isnan(prediction_sequence_list[index]):
                value = self._predict(np.array(big_bunch))
                prediction_sequence_list[index] = value

            index = index + 1
            big_bunch = []
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", UserWarning)
                for col in range(len(processed_dataframe.columns.values)):
                    if list(processed_dataframe.columns.values)[col] == self._data._dependent_variable:
                        big_bunch.append(prediction_sequence_list[index - self._seq_len:(index - self._seq_len+self._seq_len)])
                    else:
                        big_bunch.append(processed_dataframe_transform[:, col][index - self._seq_len:(index - self._seq_len+self._seq_len)])

        transformed_results = prediction_sequence_list
        if self._data._column_transforms_mapping.get(self._data._dependent_variable):
            for transform in self._data._column_transforms_mapping.get(self._data._dependent_variable):
                transformed_results = transform.inverse_transform(np.array(transformed_results).reshape(-1, 1))
                transformed_results = transformed_results.squeeze(1)

        orig_dataframe[self._data._dependent_variable + "_results"] = transformed_results
        if prediction_type == "dataframe":
            return orig_dataframe

        return orig_dataframe.spatial.to_featurelayer(output_layer_name, gis)

    def score(self):
        """
        :returns R2 score for regression model and Accuracy for classification model.
        """

        self._check_requisites()
        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")

        model = self.learn.model

        model.eval()

        dl = self.learn.data.valid_dl

        targets = []
        predictions = []
        for i in range(len(dl.x.items)):
            prediction = self._predict(dl.x.items[i])
            target = dl.y.items[i]
            targets.append(target)
            predictions.append(prediction)

        transformed_results = targets
        if self._data._column_transforms_mapping.get(self._data._dependent_variable):
            for transform in self._data._column_transforms_mapping.get(self._data._dependent_variable):
                transformed_results = transform.inverse_transform(np.array(transformed_results).reshape(-1, 1))
                transformed_results = transformed_results.squeeze(1)
        targets = transformed_results

        transformed_results = predictions
        if self._data._column_transforms_mapping.get(self._data._dependent_variable):
            for transform in self._data._column_transforms_mapping.get(self._data._dependent_variable):
                transformed_results = transform.inverse_transform(np.array(transformed_results).reshape(-1, 1))
                transformed_results = transformed_results.squeeze(1)
        predictions = transformed_results

        if self._data._is_classification:
            return (np.array(predictions)==np.array(targets)).mean()
        else:
            targets = torch.tensor(np.array(targets, dtype='float64')).to(self._device)
            predictions = torch.tensor(np.array(predictions, dtype='float64')).to(self._device)
            return float(r2_score(predictions, targets))

    def show_results(self, rows=5):
        """
        Prints the graph with predictions.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional Integer.
                                Number of rows to print.
        =====================   ===========================================
        """
        self._check_requisites()

        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")

        model = self.learn.model

        model.eval()

        dl = self.learn.data.valid_dl

        targets = []
        predictions = []
        sequence = []
        for i in range(len(dl.x.items)):
            prediction = self._predict(dl.x.items[i])
            target = dl.y.items[i]
            targets.append(target)
            predictions.append(prediction)
            sequence.append(dl.x.items[i])

        targets = np.array(targets, dtype='float64')
        predictions = np.array(predictions, dtype='float64')

        transformed_results = targets
        if self._data._column_transforms_mapping.get(self._data._dependent_variable):
            for transform in self._data._column_transforms_mapping.get(self._data._dependent_variable):
                transformed_results = transform.inverse_transform(np.array(transformed_results).reshape(-1, 1))
                transformed_results = transformed_results.squeeze(1)

        targets_inversed = transformed_results

        transformed_results = predictions
        if self._data._column_transforms_mapping.get(self._data._dependent_variable):
            for transform in self._data._column_transforms_mapping.get(self._data._dependent_variable):
                transformed_results = transform.inverse_transform(np.array(transformed_results).reshape(-1, 1))
                transformed_results = transformed_results.squeeze(1)

        predictions_inversed = transformed_results

        column_transforms_mapping = self._data._column_transforms_mapping.copy()
        # del column_transforms_mapping[self._data._dependent_variable]
        sequence_inversed = []
        keys = list(column_transforms_mapping.keys())
        for seq in sequence:
            seq_inverse = []
            index = 0
            for col_data in seq:
                transformed_data = col_data
                if len(keys) > index:
                    for transform in column_transforms_mapping.get(keys[index]):
                        transformed_data = transform.inverse_transform(np.array(transformed_data).reshape(-1, 1))
                        transformed_data = transformed_data.squeeze(1)

                seq_inverse.append(transformed_data)
                index = index + 1

            sequence_inversed.append(seq_inverse)

        if self._data._index_seq is not None:
            validation_index_seq = self._data._index_seq.take(self._data._validation_indexes_ts, axis=0)
        else:
            validation_index_seq = None

        import matplotlib.pyplot as plt
        n_items = rows
        if n_items > len(targets_inversed):
            n_items = len(targets_inversed)

        rows = int(n_items)

        fig, axs = plt.subplots(rows, 2, figsize=(10, 10))
        fig.suptitle('Ground truth vs Predictions\n\n', fontsize=16)

        for i in range(rows):
            for seq_plot in sequence_inversed[i]:
                if self._data._index_seq is not None:
                    axs[i, 0].plot(validation_index_seq[i], seq_plot)
                    axs[i, 1].plot(validation_index_seq[i], seq_plot)
                else:
                    axs[i, 0].plot(seq_plot)
                    axs[i, 1].plot(seq_plot)

                axs[i, 0].tick_params(axis="x", labelrotation=60)
                axs[i, 1].tick_params(axis="x", labelrotation=60)

            axs[i, 0].set_title(targets_inversed[i])
            axs[i, 1].set_title(predictions_inversed[i])

        plt.tight_layout()
        plt.show()


import os
from ._store import *
from .exceptions import *


class AssignmentManager(object):
    """
    This manages the assignments in the project.
    It can be accessed from the project as :py:attr:`~arcgis.apps.workforce.Project.assignments`

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    project                Required :class:`~arcgis.apps.workforce.Project`. The project to
                           manage.
    ==================     ====================================================================

    .. code-block:: python

        # Add / get assignments using assignment manager

        import arcgis
        gis = arcgis.gis.GIS("https://arcgis.com", "<username>", "<password>")
        item = gis.content.get("<item-id>")
        project = arcgis.apps.workforce.Project(item)
        assignment = project.assignments.add(assignment_type=type_1, location="100 Commercial St",
                                            geometry={'x': -7820308, 'y': 5412450}, status=0, priority=0)
        assignment_2 = project.assignments.get(object_id=2)


    """

    def __init__(self, project):
        self.project = project

    def get(self, object_id=None, global_id=None):
        """
        Gets the identified assignment by either an object id or global id.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        object_id              Optional :class:`integer`. The object id of the assignment to get
        ------------------     --------------------------------------------------------------------
        global_id              Optional :class:`string`. The global id of the assignment to get.
        ==================     ====================================================================

        :return: :class:`~arcgis.apps.workforce.Assignment`
        """
        return get_assignment(self.project, object_id, global_id)

    def search(self, where='1=1'):
        """
        Searches the assignments in the project.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        where                  Optional :class:`string`. The where clause to use to query
                               assignments. Defaults to '1=1'
        ==================     ====================================================================

        :return: :class:`List` of :class:`~arcgis.apps.workforce.Assignment`
        """
        return query_assignments(self.project, where)

    def batch_add(self, assignments):
        """
        Adds the list of assignments to the project.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        assignments            Required :class:`List` of :class:`~arcgis.apps.workforce.Assignment`.
                               The list of assignments to add.
        ==================     ====================================================================

        :return: :class:`List` of :class:`~arcgis.apps.workforce.Assignment`
        """
        return add_assignments(self.project, assignments)

    def add(self, feature=None, geometry=None, assignment_type=None,
            assigned_date=None, assignment_read=None, completed_date=None, declined_comment=None,
            declined_date=None, description=None, dispatcher=None, due_date=None, in_progress_date=None,
            location=None, notes=None, paused_date=None, priority="none", status=None,
            work_order_id=None, worker=None):
        """
        Creates and adds a new assignment to the project

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        feature                Optional :class:`~arcgis.features.Feature`.
                               A feature containing the assignments attributes. If this is provided
                               the other parameters are all ignored.
        ------------------     --------------------------------------------------------------------
        geometry               Optional :class:`Dict`.
                               A dictionary containing the assignment geometry
        ------------------     --------------------------------------------------------------------
        assignment_type        Optional :class:`~arcgis.apps.workforce.AssignmentType`.
                               The assignment type that represents this assignment.
        ------------------     --------------------------------------------------------------------
        assigned_date          Optional :class:`Date`
                               The date and time the assignment was assigned
        ------------------     --------------------------------------------------------------------
        assignment_read        Optional :class:`Bool`.
                               A flag indicating that the mobile worker has seen the assignment
        ------------------     --------------------------------------------------------------------
        completed_date         Optional :class:`Date`.
                               The date the assignment was completed
        ------------------     --------------------------------------------------------------------
        declined_comment       Optional :class:`String`.
                               The comment submitted by the mobile worker.
        ------------------     --------------------------------------------------------------------
        declined_date          Optional :class:`Date`.
                               The date the assignment was declined.
        ------------------     --------------------------------------------------------------------
        description            Optional :class:`Description`.
                               The description associated with the assignment.
        ------------------     --------------------------------------------------------------------
        dispatcher             Optional :class:`~arcgis.apps.workforce.Dispatcher`.
                               The dispatcher that assigned/created the assignment.
        ------------------     --------------------------------------------------------------------
        due_date               Optional :class:`Date`.
                               The date the assignment is due.
        ------------------     --------------------------------------------------------------------
        in_progress_date       Optional :class:`Date`.
                               The date the assignment was started.
        ------------------     --------------------------------------------------------------------
        location               Optional :class:`String`.
                               The location or address of the assignment.
        ------------------     --------------------------------------------------------------------
        notes                  Optional :class:`String`.
                               The notes associated with the assignment.
        ------------------     --------------------------------------------------------------------
        paused_date            Optional :class:`Date`.
                               The date and time the assignment was paused.
        ------------------     --------------------------------------------------------------------
        priority               Optional :class:`String`.
                               The priority of the assignment
        ------------------     --------------------------------------------------------------------
        status                 Optional :class:`String`.
                               The status of the assignment.
        ------------------     --------------------------------------------------------------------
        work_order_id          Optional :class:`String`.
                               The work order id associated with the assignment.
        ------------------     --------------------------------------------------------------------
        worker                 Optional :class:`~arcgis.apps.workforce.Worker`.
                               The worker assigned to the assignment
        ==================     ====================================================================

        :return: :class:`~arcgis.apps.workforce.Assignment`
        """
        return add_assignment(self.project, feature, geometry, assignment_type,
        assigned_date, assignment_read, completed_date, declined_comment,
        declined_date, description, dispatcher, due_date, in_progress_date,
        location, notes, paused_date, priority, status,
        work_order_id, worker)

    def batch_update(self, assignments):
        """
          Updates the list of assignments in the project.


          ==================     ====================================================================
          **Argument**           **Description**
          ------------------     --------------------------------------------------------------------
          assignments            Required :class:`List` of :class:`~arcgis.apps.workforce.Assignment`.
                                 The list of assignments to update.
          ==================     ====================================================================

          :return: :class:`List` of :class:`~arcgis.apps.workforce.Assignment`
        """
        return update_assignments(self.project, assignments)

    def batch_delete(self, assignments):
        """
          Removes the list of assignments from the project.


          ==================     ====================================================================
          **Argument**           **Description**
          ------------------     --------------------------------------------------------------------
          assignments            Required :class:`List` of :class:`~arcgis.apps.workforce.Assignment`.
                                 The list of assignments to remove.
          ==================     ====================================================================
        """
        return delete_assignments(self.project, assignments)


class AssignmentTypeManager:
    """
    This manages the assignment types in the project.
    It can be accessed from the project as :py:attr:`~arcgis.apps.workforce.Project.assignment_types`

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    project                Required :class:`~arcgis.apps.workforce.Project`. The project to
                           manage.
    ==================     ====================================================================

    .. code-block:: python

        # Add / get assignment types using assignment type manager

        import arcgis
        gis = arcgis.gis.GIS("https://arcgis.com", "<username>", "<password>")
        item = gis.content.get("<item-id>")
        project = arcgis.apps.workforce.Project(item)
        assignment_type = project.assignment_types.add(name="Tree Inspection")
        assignment_type_2 = project.assignment_types.get(name="Fix Sign")


    """

    def __init__(self, project):
        self.project = project

    def get(self, code=None, name=None):
        """
        Gets the identified assignment type by either the name or code.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        code                   Optional :class:`integer`. The code of the assignment type.
        ------------------     --------------------------------------------------------------------
        name                   Optional :class:`string`. The name of the assignment type.
        ==================     ====================================================================

        :return: :class:`~arcgis.apps.workforce.AssignmentType`
        """
        if not self.project._is_v2_project:
            return get_assignment_type(self.project, code=code, name=name)
        else:
            return get_assignment_type_v2(self.project, code=code, name=name)

    def search(self):
        """
         Gets all of the assignment types in the project.

         :return: :class:`List` of :class:`~arcgis.apps.workforce.AssignmentType`
         """
        if not self.project._is_v2_project:
            return get_assignment_types(self.project)
        else:
            return get_assignment_types_v2(self.project)

    def add(self, coded_value=None, name=None):
        """
        Adds an assignment type to the project.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        coded_value            Optional :class:`dict`. The dictionary storing the code and
                               name of the type. Only works for v1 projects.
        ------------------     --------------------------------------------------------------------
        name                   Optional :class:`String`. The name of the assignment type.
        ==================     ====================================================================

        :return: :class:`~arcgis.apps.workforce.AssignmentType`
        """
        if not self.project._is_v2_project:
            return add_assignment_type(self.project, coded_value=coded_value, name=name)
        else:
            return add_assignment_type_v2(self.project, name=name)

    def batch_add(self, assignment_types):
        """
         Adds the list of assignment types to the project.

         ==================     ====================================================================
         **Argument**           **Description**
         ------------------     --------------------------------------------------------------------
         assignment_types       Required :class:`List` of :class:`~arcgis.apps.workforce.AssignmentTypes`.
                                The list of assignment types to add.
         ==================     ====================================================================

         :return: :class:`List` of :class:`~arcgis.apps.workforce.AssignmentTypes`
         """
        if not self.project._is_v2_project:
            return add_assignment_types(self.project, assignment_types)
        else:
            return add_assignment_types_v2(self.project, assignment_types)

    def batch_update(self, assignment_types):
        """
         Updates the list of assignment types to the project.


         ==================     ====================================================================
         **Argument**           **Description**
         ------------------     --------------------------------------------------------------------
         assignment_types       Required :class:`List` of :class:`~arcgis.apps.workforce.AssignmentTypes`.
                                The list of assignment types to update.
         ==================     ====================================================================

         :return: :class:`List` of :class:`~arcgis.apps.workforce.AssignmentType`
         """
        if not self.project._is_v2_project:
            return update_assignment_types(self.project, assignment_types)
        else:
            return update_assignment_types_v2(self.project, assignment_types)

    def batch_delete(self, assignment_types):
        """
         Removes the list of assignment types to the project.


         ==================     ====================================================================
         **Argument**           **Description**
         ------------------     --------------------------------------------------------------------
         assignment_types       Required :class:`List` of :class:`~arcgis.apps.workforce.AssignmentTypes`.
                                The list of assignment types to remove.
         ==================     ====================================================================
         """
        if not self.project._is_v2_project:
            return delete_assignment_types(self.project, assignment_types)
        else:
            return delete_assignment_types_v2(self.project, assignment_types)


class AssignmentAttachmentManager(object):
    """
    This manages the attachments associated with an assignment.
    It can be accessed from the assignment as :py:attr:`~arcgis.apps.workforce.Assignment.attachments`

    ==================     ====================================================================
    **Argument**           **Description**
    ------------------     --------------------------------------------------------------------
    assignment             Required :class:`~arcgis.apps.workforce.Assignment`. The assignment to
                           manage.
    ==================     ====================================================================

    """

    def __init__(self, assignment):
        self.assignment = assignment

    def get(self):
        """
        This gets all of the Attachments belonging to the assignment.

        :return: :class:`List` of :class:`~arcgis.apps.workforce.Attachment`
        """
        return get_attachments(self.assignment)

    def add(self, file_path):
        """
         Adds the file as an attachment to the assignment.


         ==================     ====================================================================
         **Argument**           **Description**
         ------------------     --------------------------------------------------------------------
         file_path              Required :class:`string` The file to upload.
         ==================     ====================================================================
         """
        return add_attachment(self.assignment, file_path)

    def batch_delete(self, attachments):
        """
         Removes the list of attachments from the assignment.


         ==================     ====================================================================
         **Argument**           **Description**
         ------------------     --------------------------------------------------------------------
         attachments            Required :class:`List` of :class:`~arcgis.apps.workforce.Attachment`.
                                The list of attachments to delete.
         ==================     ====================================================================
         """
        return delete_attachments(self.assignment, attachments)

    def download(self, out_folder=None):
        """
         Downloads all of an assignments attachments.


         ==================     ====================================================================
         **Argument**           **Description**
         ------------------     --------------------------------------------------------------------
         out_folder             Required :class:`string`. The folder to download the attachments to.
         ==================     ====================================================================

         :return: A :class:`List` of file path strings
         """
        if not out_folder:
            out_folder = os.getcwd()
        attachments = get_attachments(self.assignment)
        paths = []
        for attachment in attachments:
            paths.append(attachment.download(out_folder))
        return paths


class DispatcherManager:
    """
     This manages the dispatchers in the project.
     It can be accessed from the project as :py:attr:`~arcgis.apps.workforce.Project.dispatchers`

     ==================     ====================================================================
     **Argument**           **Description**
     ------------------     --------------------------------------------------------------------
     project                Required :class:`~arcgis.apps.workforce.Project`. The project to
                            manage.
     ==================     ====================================================================
     
    .. code-block:: python

        # Add / get dispatchers using dispatcher manager

        import arcgis
        gis = arcgis.gis.GIS("https://arcgis.com", "<username>", "<password>")
        item = gis.content.get("<item-id>")
        project = arcgis.apps.workforce.Project(item)
        dispatcher = project.dispatchers.add(name="New Dispatcher",user_id="dispatcher_username")
        dispatcher_2 = project.dispatchers.get(user_id="dispatcher2_username")


    """

    def __init__(self, project):
        self.project = project

    def get(self, object_id=None, global_id=None, user_id=None):
        """
         This gets a dispatcher by their object id, global id, or user id.

         ==================     ====================================================================
         **Argument**           **Description**
         ------------------     --------------------------------------------------------------------
         object_id              Optional :class:`integer`. The object id of the dispatcher to get
         ------------------     --------------------------------------------------------------------
         global_id              Optional :class:`string`. The global id of the dispatcher to get.
         ------------------     --------------------------------------------------------------------
         user_id                Optional :class:`string`. The user id of the dispatcher to get.
         ==================     ====================================================================

         :return: :class:`~arcgis.apps.workforce.Dispatcher`

        """
        return get_dispatcher(self.project, object_id, global_id, user_id)

    def search(self, where='1=1'):
        """
        Searches the dispatchers in the project.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        where                  Optional :class:`string`. The where clause to use to query
                               dispatchers. Defaults to '1=1'
        ==================     ====================================================================

        :return: :class:`List` of :class:`~arcgis.apps.workforce.Dispatcher`
        """
        return query_dispatchers(self.project, where)

    def add(self, feature=None, contact_number=None, name=None, user_id=None):
        """
        Creates and adds a dispatcher to the project.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        project                Required :class:`~arcgis.apps.workforce.Project`. The project that
                               the dispatcher belongs to.
        ------------------     --------------------------------------------------------------------
        feature                Optional :class:`~arcgis.features.Feature`. The feature representing
                               the dispatcher.
        ------------------     --------------------------------------------------------------------
        contact_number         Optional :class:`String`. The contact number of the dispatcher
        ------------------     --------------------------------------------------------------------
        name                   Optional :class:`String`. The name of the dispatcher
        ------------------     --------------------------------------------------------------------
        user_id                Optional :class:`String`. The user id of the dispatcher
        ==================     ====================================================================

        :return: :class:`~arcgis.apps.workforce.Dispatcher`
        """
        return add_dispatcher(self.project, feature, contact_number, name, user_id)

    def batch_add(self, dispatchers):
        """
        Adds the list of dispatchers to the project.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        dispatchers            Required :class:`List` of :class:`~arcgis.apps.workforce.Dispatcher`.
                               The list of dispatchers to add.
        ==================     ====================================================================

        :return: :class:`List` of :class:`~arcgis.apps.workforce.Dispatcher`
        """
        return add_dispatchers(self.project, dispatchers)

    def batch_update(self, dispatchers):
        """
        Adds the list of dispatchers to update in the project.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        dispatchers            Required :class:`List` of :class:`~arcgis.apps.workforce.Dispatcher`.
                               The list of dispatchers to update.
        ==================     ====================================================================

        :return: :class:`List` of :class:`~arcgis.apps.workforce.Dispatcher`
        """
        return update_dispatchers(self.project, dispatchers)

    def batch_delete(self, dispatchers):
        """
        Removes the list of dispatchers to remove from the project.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        dispatchers            Required :class:`List` of :class:`~arcgis.apps.workforce.Dispatcher`.
                               The list of dispatchers to remove.
        ==================     ====================================================================
        """
        return delete_dispatchers(self.project, dispatchers)


class TrackManager:
    """
     This manages the tracks in the project.
     It can be accessed from the project as :py:attr:`~arcgis.apps.workforce.Project.tracks`

     ==================     ====================================================================
     **Argument**           **Description**
     ------------------     --------------------------------------------------------------------
     project                Required :class:`~arcgis.apps.workforce.Project`. The project to
                            manage.
     ==================     ====================================================================
    """
    def __init__(self, project):
        if not project._supports_tracks:
            raise WorkforceError("This Workforce Project does not support tracks.")
        self.project = project

    def get(self, object_id=None, global_id=None):
        """
         This gets a track by their object id or global id.

         ==================     ====================================================================
         **Argument**           **Description**
         ------------------     --------------------------------------------------------------------
         object_id              Optional :class:`integer`. The object id of the track to get
         ------------------     --------------------------------------------------------------------
         global_id              Optional :class:`string`. The global id of the track to get.
         ==================     ====================================================================

         :return: :class:`~arcgis.apps.workforce.Track`

        """
        return get_track(self.project, object_id, global_id)

    def search(self, where='1=1'):
        """
        Searches the tracks in the project.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        where                  Optional :class:`string`. The where clause to use to query
                               tracks. Defaults to '1=1'
        ==================     ====================================================================

        :return: :class:`List` of :class:`~arcgis.apps.workforce.Track`
        """
        return query_tracks(self.project, where)

    def add(self, feature=None, geometry=None, accuracy=None):
        """
        Adds a track to the project.

        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        feature                Optional :class:`~arcgis.features.Feature`.
                               The feature to use.
        ------------------     --------------------------------------------------------------------
        geometry               Optional :class:`Dict`.
                               A dictionary containing the assignment geometry
        ------------------     --------------------------------------------------------------------
        accuracy               Optional :class:`Float`.
                               The accuracy to use.
        ==================     ====================================================================

        :return: :class:`~arcgis.apps.workforce.Track`
        """
        return add_track(self.project, feature, geometry, accuracy)

    def batch_add(self, tracks):
        """
        Adds the list of tracks to the project.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        tracks                 Required :class:`List` of :class:`~arcgis.apps.workforce.Track`.
                               The list of tracks to add.
        ==================     ====================================================================

        :return: :class:`List` of :class:`~arcgis.apps.workforce.Track`
        """
        return add_tracks(self.project, tracks)

    def batch_delete(self, tracks):
        """
        Removes the list of tracks to remove from the project.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        tracks                 Required :class:`List` of :class:`~arcgis.apps.workforce.Track`.
                               The list of tracks to remove.
        ==================     ====================================================================
        """
        return delete_tracks(self.project, tracks)

    def batch_update(self, tracks):
        """
        Updates the list of tracks in the project.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        tracks                 Required :class:`List` of :class:`~arcgis.apps.workforce.Track`.
                               The list of tracks to update.
        ==================     ====================================================================

        :return: :class:`List` of :class:`~arcgis.apps.workforce.Track`
        """
        return update_tracks(self.project, tracks)

    @property
    def enabled(self):
        """Gets/sets if tracking is enabled for the project"""
        return self.project._tracking_enabled

    @enabled.setter
    def enabled(self, value):
        self.project._tracking_enabled = value

    @property
    def interval(self):
        """Gets/sets the tracking interval for the project (in seconds)"""
        return self.project._tracking_interval

    @interval.setter
    def interval(self, value):
        self.project._tracking_interval = value


class WorkerManager:
    """
     This manages the workers in the project
     It can be accessed from the project as :py:attr:`~arcgis.apps.workforce.Project.workers`

     ==================     ====================================================================
     **Argument**           **Description**
     ------------------     --------------------------------------------------------------------
     project                Required :class:`~arcgis.apps.workforce.Project`. The project to
                            manage.
     ==================     ====================================================================

    .. code-block:: python

        # Add / get workers using worker manager

        import arcgis
        gis = arcgis.gis.GIS("https://arcgis.com", "<username>", "<password>")
        item = gis.content.get("<item-id>")
        project = arcgis.apps.workforce.Project(item)
        worker = project.workers.add(user_id="worker_username", name="Worker One", status=0)


    """
    def __init__(self, project):
        self.project = project

    def get(self, object_id=None, global_id=None, user_id=None):
        """
          This gets a worker by their object id, global id, or user id.

          ==================     ====================================================================
          **Argument**           **Description**
          ------------------     --------------------------------------------------------------------
          object_id              Optional :class:`integer`. The object id of the worker to get
          ------------------     --------------------------------------------------------------------
          global_id              Optional :class:`string`. The global id of the worker to get.
          ------------------     --------------------------------------------------------------------
          user_id                Optional :class:`string`. The user id of the worker to get.
          ==================     ====================================================================

          :return: :class:`~arcgis.apps.workforce.Worker`

         """
        return get_worker(self.project, object_id, global_id, user_id)

    def search(self, where='1=1'):
        """
        Searches the workers in the project.


        ==================     ====================================================================
        **Argument**           **Description**
        ------------------     --------------------------------------------------------------------
        where                  Optional :class:`string`. The where clause to use to query
                               workers. Defaults to '1=1'
        ==================     ====================================================================

        :return: :class:`List` of :class:`~arcgis.apps.workforce.Worker`
        """
        return query_workers(self.project, where)

    def batch_add(self, workers):
        """
          Adds the list of workers to the project.


          ==================     ====================================================================
          **Argument**           **Description**
          ------------------     --------------------------------------------------------------------
          workers                Required :class:`List` of :class:`~arcgis.apps.workforce.Worker`.
                                 The list of workers to add.
          ==================     ====================================================================

          :return: :class:`List` of :class:`~arcgis.apps.workforce.Worker`
          """
        return add_workers(self.project, workers)

    def add(self, feature=None, geometry=None, contact_number=None,
               name=None, notes=None, status="not_working", title=None, user_id=None):
        """
            Creates and adds a new worker to the project.

            ==================     ====================================================================
            **Argument**           **Description**
            ------------------     --------------------------------------------------------------------
            feature                Optional :class:`~arcgis.features.Feature`. The feature representing
                                   the worker.
            ------------------     --------------------------------------------------------------------
            geometry               Optional :class:`Dict`. The geometry of the worker.
            ------------------     --------------------------------------------------------------------
            contact_number         Optional :class:`String`. The contact number of the worker.
            ------------------     --------------------------------------------------------------------
            name                   Optional :class:`String`. The name of the worker.
            ------------------     --------------------------------------------------------------------
            notes                  Optional :class:`String`. The notes about the worker.
            ------------------     --------------------------------------------------------------------
            status                 Optional :class:`String`. The status of the worker.
            ------------------     --------------------------------------------------------------------
            title                  Optional :class:`String`. The title of the worker.
            ------------------     --------------------------------------------------------------------
            user_id                Optional :class:`String`. The user id of the worker
            ==================     ====================================================================

            :return: :class:`~arcgis.apps.workforce.Worker`
        """
        return add_worker(self.project, feature, geometry, contact_number, name, notes, status, title, user_id)

    def batch_update(self, workers):
        """
          Adds the list of workers to update in the project.


          ==================     ====================================================================
          **Argument**           **Description**
          ------------------     --------------------------------------------------------------------
          workers                Required :class:`List` of :class:`~arcgis.apps.workforce.Worker`.
                                 The list of workers to update.
          ==================     ====================================================================

          :return: :class:`List` of :class:`~arcgis.apps.workforce.Worker`
          """
        return update_workers(self.project, workers)

    def batch_delete(self, workers):
        """
          Adds the list of workers to remove from the project.


          ==================     ====================================================================
          **Argument**           **Description**
          ------------------     --------------------------------------------------------------------
          workers                Required :class:`List` of :class:`~arcgis.apps.workforce.Worker`.
                                 The list of workers to remove.
          ==================     ====================================================================
          """
        return delete_workers(self.project, workers)


class AssignmentIntegrationManager:
    """
     This manages the assignment integrations in the project
     It can be accessed from the project as :py:attr:`~arcgis.apps.workforce.Project.integrations`
    
     For a version 2 (offline-enabled project), an integration is represented by an object
     :class:`~arcgis.apps.workforce.Integration`
     object and can be updated in the same fashion as Assignment, AssignmentType, Dispatcher, Project,
     and Worker objects.
     
     For a version 1 project, an integration in Workforce consists of a formatted dictionary.
     Two examples are shown below:

     .. code-block:: python

            navigator_integration = {
                "id": "default-navigator",
                "prompt": "Navigate to Assignment",
                "urlTemplate": "arcgis-navigator://?stop=${assignment.latitude},${assignment.longitude}&stopname=${assignment.location}&callback=arcgis-workforce://&callbackprompt=Workforce"
            }

            explorer_integration = {
                "id": "default-explorer",
                "prompt": "Explore at Assignment",
                "assignmentTypes": {
                    "1": {
                        "urlTemplate": "arcgis-explorer://?itemID=651324c8661b42c897657f8afbe846qe&center=${assignment.latitude},${assignment.longitude}&scale=9000"
                    }
            }

     The urlTemplate can be generated by using the :py:mod:`~arcgis.apps.integrations` module

     ==================     ====================================================================
     **Argument**           **Description**
     ------------------     --------------------------------------------------------------------
     project                Required :class:`~arcgis.apps.workforce.Project`. The project to
                            manage.
     ==================     ====================================================================
    """
    def __init__(self, project):
        self.project = project
        # A version 2 Workforce project stores integrations in a table in its base feature layer collection
        if self.project._is_v2_project:
            self.integration_table = project.integrations_table

    def get(self, integration_id):
        """
          This gets an integration dictionary by its id

          ==================     ====================================================================
          **Argument**           **Description**
          ------------------     --------------------------------------------------------------------
          integration_id         Required :class:`string`. The id of the integration. This is field
                                 'appid' for a Version 2 Workforce project.
          ==================     ====================================================================

          :returns: Version 1: :class:`dict` or :class:`None`, Version 2: :class:`Integration`

         """
        if self.project._is_v2_project:
            return get_integration(self.project, integration_id)
        else:
            for integration in self.project._item_data["assignmentIntegrations"]:
                if integration_id == integration["id"]:
                    return integration
            return None

    def search(self, where="1=1"):
        """
            This returns all of the assignment integrations for the project
            
            ==================     ====================================================================
            **Argument**           **Description**
            ------------------     --------------------------------------------------------------------
            where                  Optional :class:`string`. ArcGIS where clause - version 2 projects
                                   only. Defaults to "1=1"
            ------------------     --------------------------------------------------------------------

            :returns: :class:`List` A list of the integrations.
        """
        if self.project._is_v2_project:
            return query_integrations(self.project, where=where)
        else:
            return self.project._item_data["assignmentIntegrations"]

    def add(self, integration_id, prompt, url_template=None, assignment_types=None):
        """
            This adds an integration to the project

            ==================     ====================================================================
            **Argument**           **Description**
            ------------------     --------------------------------------------------------------------
            integration_id         Required :class:`string`. The id of the integration
            ------------------     --------------------------------------------------------------------
            prompt                 Required: :class:`string`. The prompt to display.
            ------------------     --------------------------------------------------------------------
            url_template           Required for version 2 Workforce projects. Optional for version 1.
                                   :class:`string`. The url template that is used for app linking.
            ------------------     --------------------------------------------------------------------
            assignment_types       Optional: :class:`string` or :class:`list`
                                   Version 2 Projects:
                                   String which is a globalid representing an assignment type. This is
                                   stored at assignment_type.code. If you pass a list for this object,
                                   it will add multiple integrations, with identical integration_id,
                                   prompt, and url_template values, with differring assignment_types
                                   values.
                                   
                                   Version 1 Projects: :class:`dict`.
                                   A dictionary containing assignment type
                                   codes as keys and a dictionaries that contains a "urlTemplate" for each
                                   code as values. If provided, this will override any general url_template specified.
            ==================     ====================================================================

            :returns: :class:`dict` Version 1: dict representing the integration, Version 2: :class:`Integration`
        """
        if self.project._is_v2_project:
            if isinstance(assignment_types, list):
                integration = [add_integration(project=self.project, integration_id=integration_id, prompt=prompt, url_template=url_template, assignment_type=a) for a in assignment_types]
            else:
                integration = add_integration(project=self.project, integration_id=integration_id, prompt=prompt, url_template=url_template, assignment_type=assignment_types)
            return integration
        else:
            integration = dict()
            if integration_id:
                integration["id"] = integration_id
            if prompt:
                integration["prompt"] = prompt
            if assignment_types:
                integration["assignmentTypes"] = assignment_types
            if url_template:
                integration["urlTemplate"] = url_template
            new_integration = self._validate(integration)
            self.project._item_data["assignmentIntegrations"].append(new_integration)
            self.project._update_data()
            return new_integration

    def batch_add(self, integrations):
        """
            ==================     ====================================================================
            **Argument**           **Description**
            ------------------     --------------------------------------------------------------------
            integrations           Required :class:`List` of :class:`dict`. The integrations to add
            ==================     ====================================================================

            :returns: :class:`List` The list of integrations that were added
        """
        if self.project._is_v2_project:
            return add_integrations(self.project, integrations)
        else:
            for integration in integrations:
                new_integration = self._validate(integration)
                self.project._item_data["assignmentIntegrations"].append(new_integration)
            self.project._update_data()
            return integrations

    def batch_delete(self, integrations):
        """
            ==================     ====================================================================
            **Argument**           **Description**
            ------------------     --------------------------------------------------------------------
            integrations            Required :class:`List` of :class:`dict`. The integrations to delete
            ==================     ====================================================================
        """
        if self.project._is_v2_project:
            delete_integrations(project=self.project, integrations=integrations)
        else:
            self.project._item_data["assignmentIntegrations"] = [e for e in self.project._item_data["assignmentIntegrations"] if e not in integrations]
            self.project._update_data()

    def _validate(self, integration):
        """Validates an integration for a version 1 project before adding it. Returns integration (that may have
        been modified in this validation method"""
        if "id" not in integration:
            raise ValidationError("Assignment integration must contain an id", self)
        elif integration["id"] in [at["id"] for at in self.project._item_data["assignmentIntegrations"]]:
            raise ValidationError("Assignment integration contains duplicate id for version 1 project", self)
        if "prompt" not in integration:
            raise ValidationError("Assignment integration must contain a prompt", self)
        if "assignmentTypes" in integration:
            copy_dict = integration["assignmentTypes"].copy()
            for key, value in copy_dict.items():
                if isinstance(key, str):
                    if "urlTemplate" not in value:
                        raise ValidationError("Assignment integration must contain a urlTemplate", self)
                    if key not in [at.name for at in self.project.assignment_types.search()]:
                        raise ValidationError("Invalid assignment type in integration", self)
                    # swap the name for the code
                    integration["assignmentTypes"][self.project.assignment_types.get(name=key).code] = integration[
                        "assignmentTypes"].pop(key)
                elif isinstance(key, int):
                    if key not in [at.code for at in self.project.assignment_types.search()]:
                        raise ValidationError("Invalid assignment type in integration", self)
                    elif "urlTemplate" not in value:
                        raise ValidationError("Assignment integration must contain a urlTemplate", self)
                else:
                    raise ValidationError("Invalid assignment type", self)
        elif "urlTemplate" not in integration:
            raise ValidationError("Assignment integration must contain a urlTemplate", self)
        return integration


"""
The code is borrowed from:
-------------------------------------------------------------------------------

https://github.com/anilbatra2185/road_connectivity/blob/master/model/linknet.py

"""

import math

import torch.nn as nn
import torch.nn.functional as F

affine_par = True


class BasicResnetBlock(nn.Module):
    expansion = 1

    def __init__(self, inplanes, planes, stride=1, padding=1, downsample=None):
        super(BasicResnetBlock, self).__init__()

        self.conv1 = nn.Conv2d(
            inplanes, planes, kernel_size=3, stride=stride, padding=padding, bias=False
        )
        self.bn1 = nn.BatchNorm2d(planes)
        self.relu = nn.ReLU(inplace=True)

        self.conv2 = nn.Conv2d(
            planes, planes, kernel_size=3, stride=stride, padding=padding, bias=False
        )
        self.bn2 = nn.BatchNorm2d(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x):
        residual = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            residual = self.downsample(x)

        out += residual
        out = self.relu(out)

        return out


class DecoderBlock(nn.Module):
    def __init__(self, in_channels, n_filters, group=1):
        super(DecoderBlock, self).__init__()

        # B, C, H, W -> B, C/4, H, W
        self.conv1 = nn.Conv2d(in_channels, in_channels // 4, 1, groups=group)
        self.norm1 = nn.BatchNorm2d(in_channels // 4)
        self.relu1 = nn.ReLU(inplace=True)

        # B, C/4, H, W -> B, C/4, H, W
        self.deconv2 = nn.ConvTranspose2d(
            in_channels // 4,
            in_channels // 4,
            3,
            stride=2,
            padding=1,
            output_padding=1,
            groups=group,
        )
        self.norm2 = nn.BatchNorm2d(in_channels // 4)
        self.relu2 = nn.ReLU(inplace=True)

        # B, C/4, H, W -> B, C, H, W
        self.conv3 = nn.Conv2d(in_channels // 4, n_filters, 1, groups=group)
        self.norm3 = nn.BatchNorm2d(n_filters)
        self.relu3 = nn.ReLU(inplace=True)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            if isinstance(m, nn.ConvTranspose2d):
                n = m.kernel_size[0] * m.kernel_size[1] * m.out_channels
                m.weight.data.normal_(0, math.sqrt(2.0 / n))
            elif isinstance(m, nn.BatchNorm2d):
                m.weight.data.fill_(1)
                m.bias.data.zero_()

    def forward(self, x):
        x = self.conv1(x)
        x = self.norm1(x)
        x = self.relu1(x)
        x = self.deconv2(x)
        x = self.norm2(x)
        x = self.relu2(x)
        x = self.conv3(x)
        x = self.norm3(x)
        x = self.relu3(x)
        return x


class HourglassModuleMTL(nn.Module):
    def __init__(self, block, num_blocks, planes, depth):
        super(HourglassModuleMTL, self).__init__()
        self.depth = depth
        self.block = block
        self.upsample = nn.Upsample(scale_factor=2)
        self.hg = self._make_hour_glass(block, num_blocks, planes, depth)

    def _make_residual1(self, block, num_blocks, planes):
        layers = []
        for i in range(0, num_blocks):
            layers.append(block(planes * block.expansion, planes))
        return nn.Sequential(*layers)

    def _make_hour_glass(self, block, num_blocks, planes, depth):
        hg = []
        for i in range(depth):
            res = []
            for j in range(4):
                res.append(self._make_residual1(block, num_blocks, planes))
            if i == 0:
                res.append(self._make_residual1(block, num_blocks, planes))
                res.append(self._make_residual1(block, num_blocks, planes))
            hg.append(nn.ModuleList(res))
        return nn.ModuleList(hg)

    def _hour_glass_forward(self, n, x):
        rows = x.size(2)
        cols = x.size(3)

        up1 = self.hg[n - 1][0](x)
        low1 = F.max_pool2d(x, 2, stride=2, ceil_mode=True)
        low1 = self.hg[n - 1][1](low1)

        if n > 1:
            low2_1, low2_2 = self._hour_glass_forward(n - 1, low1)
        else:
            low2_1 = self.hg[n - 1][4](low1)
            low2_2 = self.hg[n - 1][5](low1)
        low3_1 = self.hg[n - 1][2](low2_1)
        low3_2 = self.hg[n - 1][3](low2_2)
        up2_1 = self.upsample(low3_1)
        up2_2 = self.upsample(low3_2)
        out_1 = up1 + up2_1[:, :, :rows, :cols]
        out_2 = up1 + up2_2[:, :, :rows, :cols]

        return out_1, out_2

    def forward(self, x):
        return self._hour_glass_forward(self.depth, x)


class StackHourglassMultiTaskModel(nn.Module):
    def __init__(
        self, encoder=None, task1_classes=2, task2_classes=37, chip_size=None, n_bands=3, **kwargs
    ):
        super(StackHourglassMultiTaskModel, self).__init__()

        num_stacks = kwargs.get("num_stacks", 2)
        num_blocks = kwargs.get("num_blocks", 1)
        hg_num_blocks = kwargs.get("hg_num_blocks", 3)
        depth = kwargs.get("depth", 3)
        block = BasicResnetBlock

        self.chip_size = chip_size
        self.inplanes = 64
        self.num_feats = 128
        self.num_stacks = num_stacks
        self.conv1 = nn.Conv2d(
            n_bands, self.inplanes, kernel_size=7, stride=2, padding=3, bias=True
        )
        self.bn1 = nn.BatchNorm2d(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.layer1 = self._make_residual(block, self.inplanes, 1)
        self.layer2 = self._make_residual(block, self.inplanes, num_blocks)
        self.layer3 = self._make_residual(block, self.num_feats, num_blocks)
        self.maxpool = nn.MaxPool2d(2, stride=2, ceil_mode=True)

        # build hourglass modules
        ch = self.num_feats * block.expansion
        hg = []
        res_1, fc_1, score_1, _fc_1, _score_1 = [], [], [], [], []
        res_2, fc_2, score_2, _fc_2, _score_2 = [], [], [], [], []

        for i in range(num_stacks):
            hg.append(HourglassModuleMTL(block, hg_num_blocks, self.num_feats, depth))

            res_1.append(self._make_residual(block, self.num_feats, hg_num_blocks))
            res_2.append(self._make_residual(block, self.num_feats, hg_num_blocks))

            fc_1.append(self._make_fc(ch, ch))
            fc_2.append(self._make_fc(ch, ch))

            score_1.append(nn.Conv2d(ch, task1_classes, kernel_size=1, bias=True))
            score_2.append(nn.Conv2d(ch, task2_classes, kernel_size=1, bias=True))
            if i < num_stacks - 1:
                _fc_1.append(nn.Conv2d(ch, ch, kernel_size=1, bias=True))
                _fc_2.append(nn.Conv2d(ch, ch, kernel_size=1, bias=True))
                _score_1.append(nn.Conv2d(task1_classes, ch, kernel_size=1, bias=True))
                _score_2.append(nn.Conv2d(task2_classes, ch, kernel_size=1, bias=True))

        self.hg = nn.ModuleList(hg)
        self.res_1 = nn.ModuleList(res_1)
        self.fc_1 = nn.ModuleList(fc_1)
        self.score_1 = nn.ModuleList(score_1)
        self._fc_1 = nn.ModuleList(_fc_1)
        self._score_1 = nn.ModuleList(_score_1)

        self.res_2 = nn.ModuleList(res_2)
        self.fc_2 = nn.ModuleList(fc_2)
        self.score_2 = nn.ModuleList(score_2)
        self._fc_2 = nn.ModuleList(_fc_2)
        self._score_2 = nn.ModuleList(_score_2)

        # Final Classifier
        self.decoder1 = DecoderBlock(self.num_feats, self.inplanes)
        # self.decoder1_score = nn.Conv2d(
        #     self.inplanes, task1_classes, kernel_size=1, bias=True
        # )
        self.finaldeconv1 = nn.ConvTranspose2d(self.inplanes, 32, 3, stride=2)
        self.finalrelu1 = nn.ReLU(inplace=True)
        self.finalconv2 = nn.Conv2d(32, 32, 3)
        self.finalrelu2 = nn.ReLU(inplace=True)
        self.finalconv3 = nn.Conv2d(32, task1_classes, 2, padding=1)

        # Final Classifier
        self.angle_decoder1 = DecoderBlock(self.num_feats, self.inplanes)
        # self.angle_decoder1_score = nn.Conv2d(
        #     self.inplanes, task2_classes, kernel_size=1, bias=True
        # )
        self.angle_finaldeconv1 = nn.ConvTranspose2d(self.inplanes, 32, 3, stride=2)
        self.angle_finalrelu1 = nn.ReLU(inplace=True)
        self.angle_finalconv2 = nn.Conv2d(32, 32, 3)
        self.angle_finalrelu2 = nn.ReLU(inplace=True)
        self.angle_finalconv3 = nn.Conv2d(32, task2_classes, 2, padding=1)

    def _make_residual(self, block, planes, blocks, stride=1):
        downsample = None
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                nn.Conv2d(
                    self.inplanes, planes * block.expansion,
                    kernel_size=1, stride=stride, bias=True,
                )
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample=downsample))
        self.inplanes = planes * block.expansion
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes))

        return nn.Sequential(*layers)

    def _make_fc(self, inplanes, outplanes):
        bn = nn.BatchNorm2d(inplanes)
        conv = nn.Conv2d(inplanes, outplanes, kernel_size=1, bias=True)
        return nn.Sequential(conv, bn, self.relu)

    def forward(self, x):
        # out_1 = []
        # out_2 = []

        rows = x.size(2)
        cols = x.size(3)

        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)

        x = self.layer1(x)
        x = self.maxpool(x)
        x = self.layer2(x)
        x = self.layer3(x)
        for i in range(self.num_stacks):
            y1, y2 = self.hg[i](x)
            y1, y2 = self.res_1[i](y1), self.res_2[i](y2)
            y1, y2 = self.fc_1[i](y1), self.fc_2[i](y2)
            score1, score2 = self.score_1[i](y1), self.score_2[i](y2)
            # out_1.append(
            #     score1[:, :, : int(math.ceil(rows / 4.0)), : int(math.ceil(cols / 4.0))]
            # )
            # out_2.append(
            #     score2[:, :, : int(math.ceil(rows / 4.0)), : int(math.ceil(cols / 4.0))]
            # )
            if i < self.num_stacks - 1:
                _fc_1, _fc_2 = self._fc_1[i](y1), self._fc_2[i](y2)
                _score_1, _score_2 = self._score_1[i](score1), self._score_2[i](score2)
                x = x + _fc_1 + _score_1 + _fc_2 + _score_2

        # Final Classification
        d1 = self.decoder1(y1)[
            :, :, : int(math.ceil(rows / 2.0)), : int(math.ceil(cols / 2.0))
        ]
        # d1_score = self.decoder1_score(d1)
        # out_1.append(d1_score)
        f1 = self.finaldeconv1(d1)
        f2 = self.finalrelu1(f1)
        f3 = self.finalconv2(f2)
        f4 = self.finalrelu2(f3)
        f5 = self.finalconv3(f4)
        # out_1.append(f5)


            # Final Classification
        a_d1 = self.angle_decoder1(y2)[
                :, :, : int(math.ceil(rows / 2.0)), : int(math.ceil(cols / 2.0))
               ]
        # a_d1_score = self.angle_decoder1_score(a_d1)
        # out_2.append(a_d1_score)
        a_f1 = self.angle_finaldeconv1(a_d1)
        a_f2 = self.angle_finalrelu1(a_f1)
        a_f3 = self.angle_finalconv2(a_f2)
        a_f4 = self.angle_finalrelu2(a_f3)
        a_f5 = self.angle_finalconv3(a_f4)
        # out_2.append(a_f5)

        return f5, a_f5
        #else:
        #    return f5



import uuid
import arcgis
from .._utils._basewidget import _BaseWidget
from .._utils._basewidget import NoDataProperties


class List(_BaseWidget):
    """
    Creates a dashboard List widget.

    =========================   ===========================================
    **Argument**                **Description**
    -------------------------   -------------------------------------------
    item                        Required Portal Item object. Item object can
                                be a Feature Layer or a MapWidget.
    -------------------------   -------------------------------------------
    name                        Optional string. Name of the List widget.
    -------------------------   -------------------------------------------
    layer                       Optional integer. Layer number when item is
                                a mapwidget.
    -------------------------   -------------------------------------------
    title                       Optional string. Title or Caption for the
                                widget.
    -------------------------   -------------------------------------------
    description                 Optional string. Description for the widget.
    =========================   ===========================================
    """
    def __init__(self, item, name='List', layer=0, title=None, description=None):
        super().__init__(name, title, description)

        if item.type not in ['Feature Service', 'mapWidget']:
            raise Exception("Please specify an item")

        self.item = item
        self.type = "listWidget"
        self.layer = layer

        self._seperator_color = '#f0f0f0'
        self._selection_color = '#0f0f0f'
        self._selection_text_color = '#123210'
        
        self._max_features = 25
        self._selection_mode = "single"
        
        self._show_last_update = True

        self._nodata = NoDataProperties._nodata_init()

        # List Block
        self._list_text = None
        self._list_icon = "symbol"
        self._events = Events._create_events()

    @classmethod
    def _from_json(cls, widget_json):
        gis = arcgis.env.active_gis
        itemid = widget_json["datasets"]["datasource"]["itemid"]
        name = widget_json["name"]
        item = gis.content.get(itemid)
        title = widget_json["caption"]
        description = widget_json["description"]
        lists = List(item, name, title, description)

        return lists

    @property
    def events(self):
        """
        :return: list of events attached to the widget.
        """
        return self._events

    @property
    def max_features(self):
        """
        :return: max number of features to display.
        """
        return self._max_features

    @max_features.setter
    def max_features(self, value):
        """
        Set max number of features to display.
        """
        self._max_features = int(value)

    @property
    def show_last_update(self):
        """
        :return: show last update or not.
        """
        return self._show_last_update
    
    @show_last_update.setter
    def show_last_update(self, value):
        """
        Set show last update or not.
        """
        self._show_last_update = value
        if not isinstance(value, bool):
            self._show_last_update = True

    @property
    def seperator_color(self):
        """
        :return: seperator color.
        """
        return self._seperator_color

    @seperator_color.setter
    def seperator_color(self, value):
        """
        Set seperator color.
        """
        self._seperator_color = value
        if not isinstance(value, str):
            self._seperator_color = '#f0f0f0'
    
    @property
    def selection_color(self):
        """
        :return: selection color.
        """
        return self._selection_color

    @selection_color.setter
    def selection_color(self, value):
        """
        Set selection color.+
        """
        self._selection_color = value
        if not isinstance(value, str):
            self._selection_color = '#0f0f0f'

    @property
    def selection_text_color(self):
        """
        :return: selection text color.
        """
        return self._selection_text_color

    @selection_text_color.setter
    def selection_text_color(self, value):
        """
        Set selection text color.
        """
        self._selection_text_color = value
        if not isinstance(value, str):
            self._selection_text_color = '#123210'    
    @property
    def list_text(self):
        """
        :return: list text.
        """
        return self._list_text
    
    @list_text.setter
    def list_text(self, value):
        """
        Set list text.
        """
        self._list_text = value
        if not isinstance(value, str):
            self._list_text = ''
    
    @property
    def list_icon(self):
        """
        :return: use icon for list or not.
        """
        return self._list_icon
    
    @list_icon.setter
    def list_icon(self, value):
        """
        Set use icon for list or not.
        """
        self._list_icon = value
        if not isinstance(value, str):
            self._list_icon = "symbol"

    @property
    def selection_mode(self):
        """
        :return: selection mode.
        """
        return self._selection_mode

    @selection_mode.setter
    def selection_mode(self, value):
        """
        Set selection mode
        """
        self._selection_mode = value
        if not isinstance(value, str):
            self._selection_mode = "single"
    
    @property
    def no_data(self):
        """
        :return: Nodata Object, set various nodata properties
        """
        return self._nodata
        
    def _convert_to_json(self):
        if self.item.type == 'mapWidget':
            wlayer = self.item.layers[self.layer]
            widget_id = self.item._id
            layer_id = wlayer["id"]
            self._datasource = {"id":str(widget_id)+'#'+str(layer_id)}
        else:
            self._datasource = {
                        "type": "featureServiceDataSource",
                        "itemId": self.item.itemid,
                        "layerId": 0,
                        "table": True
                    }
        json_data = {
            "type":"listWidget",
            "iconType":self._list_icon,
            "selectionMode":self._selection_mode,
            "separatorColor":self._seperator_color,
            "selectionColor":self._selection_color,
            "selectionTextColor":self._selection_text_color,
            "events": [],
            "selectionMode": "multi",
            "datasets":[
                {
                    "type":"serviceDataset",
                    "dataSource":self._datasource,
                    "outFields":["*"],
                    "groupByFields":[],
                    "orderByFields":[],
                    "statisticDefinitions":[],
                    "maxFeatures":self._max_features,
                    "querySpatialRelationship":"esriSpatialRelIntersects",
                    "returnGeometry":False,
                    "clientSideStatistics":False,
                    "name":"main"
                }
            ],
            "id":self._id,
            "name":self.name,
            "showLastUpdate":self.show_last_update,
            "noDataVerticalAlignment":self._nodata.alignment,
            "showCaptionWhenNoData":self._nodata.show_title,
            "showDescriptionWhenNoData":self._nodata.show_description
        }
        if self.events.enable:
            json_data["events"].append({"type":self.events.type, "actions":self.events.synced_widgets})
            json_data["selectionMode"] = self.events.selection_mode

        return json_data


class Events(object):

    @classmethod
    def _create_events(cls, enable=False):
        events = Events()

        events._enable = False
        events._selection_mode = "single"
        events._type = "selectionChanged"
        events._actions = []

        events.enable = enable

        return events

    @property
    def selection_mode(self):
        """
        :return: Selection mode of events.
        """
        return self._selection_mode

    @selection_mode.setter
    def selection_mode(self, value):
        """
        Set Selection mode of events.
        """
        if value in ["single", "multi"]:
            self._selection_mode = value
        else:
            raise Exception("Please specify selection_mode from 'single' and 'multi'")
    
    @property
    def enable(self):
        """
        :return: if events are enabled or not.
        """
        return self._enable

    @enable.setter
    def enable(self, value):
        """
        Set if events are enabled or not.
        """
        self._enable = bool(value)

    @property
    def type(self):
        """
        :return: type of trigger for event.
        """
        return self._type

    @property
    def synced_widgets(self):
        """
        :return: List of synced widgets.
        """
        return self._actions

    def sync_map(self, action_type, widget):
        """
        Synchronize a mapWidget with List for triggered events.

        =========================   ===========================================
        **Argument**                **Description**
        -------------------------   -------------------------------------------
        action_type                 Required string. Actions can be one of 
                                    "zoom", "flash", "show_popup", "pan".
        -------------------------   -------------------------------------------
        widget                      Required MapWidget item. Name of the map
                                    widget.
        =========================   ===========================================
        """
        if self.enable == False:
            raise Exception("Please enable events")
        else:
            if widget.type == "mapWidget":
                if action_type in ["zoom", "flash", "show_popup", "pan"]:
                    self._actions.append({"type": action_type, "targetId": widget._id})
                elif action_type == "filter":
                    if self._targetid is not None:
                        self._actions.append({"type":action_type, "by":"whereClause", "targetId":self._targetid})
                    else:
                        raise Exception("This operation is not suitable for given dataSource.")
                else:
                    raise Exception("Please select action_type from 'zoom', 'flash', 'show_popup' and 'pan'")
            else:
                raise Exception("Please select a map widget")

    def sync_widget(self, widgets):
        """
        Synchronize non-mapWidget type widgets with List for triggered events.

        =========================   ===========================================
        **Argument**                **Description**
        -------------------------   -------------------------------------------
        widget                      Required widget item or list of widget items
                                    .Name of the widgets to be synced.
        =========================   ===========================================
        """
        if self.enable == False:
            raise Exception("Please enable events")

        else:
            if isinstance(widgets, list):
                for widget in widgets:
                    if widget.type == "mapWidget":
                        raise Exception("Use sync_map method to add actions for map widgets") ##duplicate or erase
                    else:
                        action_type = "filter"
                        widget_id = str(widget._id)+'#main'
                        self._actions.append({"type":action_type, "by":"whereClause", "targetId":widget_id})
            else:
                if widgets.type == "mapWidget":
                    raise Exception("Use sync_map method to add actions for map widgets") ##duplicate or erase
                else:
                    action_type = "filter"
                    widget_id = str(widgets._id)+'#main'
                    self._actions.append({"type":action_type, "by":"whereClause", "targetId":widget_id})




# MIT License

# Copyright (c) 2019 Hengshuang Zhao

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Based on https://github.com/hszhao/semseg

import torch
import warnings
import numpy as np
import torch.nn.functional as F
import torch.nn as nn
import torch
from torchvision import models
import math
from fastai.callbacks.hooks import hook_outputs
from fastai.vision.learner import create_body
from fastai.callbacks.hooks import model_sizes
from torchvision.models.segmentation.segmentation import _segm_resnet
from torchvision.models.segmentation.deeplabv3 import DeepLabHead, DeepLabV3
from torchvision.models.segmentation.fcn import FCNHead
from ._arcgis_model import _get_backbone_meta
from fastprogress.fastprogress import progress_bar
from ._PointRend import PointRendSemSegHead
from fastai.vision import flatten_model

class Deeplab(nn.Module):
    def __init__(self, num_classes, backbone_fn, chip_size=224, pointrend=True, keep_dilation=False):
        super().__init__()
        self.pointrend = pointrend
        self.vgg = False
        if getattr(backbone_fn, '_is_multispectral', False):
            self.backbone = create_body(backbone_fn, pretrained=True, cut=_get_backbone_meta(backbone_fn.__name__)['cut'])
        else:
            self.backbone = create_body(backbone_fn, pretrained=True)
        
        backbone_name = backbone_fn.__name__

        ## Support for different backbones
        if "densenet" in backbone_name or "vgg" in backbone_name:
            hookable_modules = list(self.backbone.children())[0]
        else:
            hookable_modules = list(self.backbone.children())
        
        if "vgg" in backbone_name:
            modify_dilation_index = -5
            self.vgg = True
        else:
            if self.pointrend and not keep_dilation:
                modify_dilation_index = -1
            else:
                modify_dilation_index = -2
            
        if backbone_name == 'resnet18' or backbone_name == 'resnet34':
            module_to_check = 'conv' 
        else:
            module_to_check = 'conv2'
        
        if "vgg" in backbone_name:
            hooks = [hookable_modules[i-1] for i, module in enumerate(hookable_modules) if isinstance(module, nn.MaxPool2d)]

        else:
            hooks = [hookable_modules[-2], hookable_modules[-4]]

        custom_idx = 0
        for i, module in enumerate(hookable_modules[modify_dilation_index:]): 
            dilation = 2 * (i + 1)
            padding = 2 * (i + 1)
            for n, m in module.named_modules():
                if module_to_check in n:
                    m.dilation, m.padding, m.stride = (dilation, dilation), (padding, padding), (1, 1)
                elif 'downsample.0' in n:
                    m.stride = (1, 1)                    
                    
            if "vgg" in backbone_fn.__name__:
                if isinstance(module, nn.Conv2d):
                    dilation = 2 * (custom_idx + 1)
                    padding = 2 * (custom_idx + 1)
                    module.dilation, module.padding, module.stride = (dilation, dilation), (padding, padding), (1, 1)
                    custom_idx += 1

        ## Hook at the index where we need to get the auxillary logits out along with Fine-grained features 
        self.hook = hook_outputs(hooks)
        
        ## returns the size of various activations
        feature_sizes = model_sizes(self.backbone, size=(chip_size, chip_size))
        
        if not self.vgg:
            ## Geting the number of channel persent in stored activation inside of the hook
            num_channels_aux_classifier = self.hook[0].stored.shape[1]
            ## Get number of channels in the last layer
            num_channels_classifier = feature_sizes[-1][1]
        else:
            num_channels_aux_classifier = self.hook[-2].stored.shape[1]
            num_channels_classifier = self.hook[-1].stored.shape[1]

        self.classifier = DeepLabHead(num_channels_classifier, num_classes)
        self.aux_classifier = FCNHead(num_channels_aux_classifier, num_classes)

        if self.pointrend:

            if self.vgg:
                num_channels = self.hook[-3].stored.shape[1] + self.hook[-4].stored.shape[1]
                stride = chip_size / self.hook[-1].stored.shape[2]  
            else:
                num_channels = self.hook[1].stored.shape[1]
                stride = chip_size / feature_sizes[-1][2]

            subdivision_steps = math.log(stride, 2)
            self.pointrend_head = PointRendSemSegHead(num_classes,
                                                    num_channels,
                                                    train_num_points=(chip_size/stride)**2,
                                                    subdivision_num_points=(chip_size/(stride/2))**2,
                                                    subdivision_steps=subdivision_steps)

    def forward(self, x):

        x_size = x.size()
        x = self.backbone(x)
        features = self.hook.stored
        
        if self.vgg:
            x = self.classifier(features[-1])
        else:
            x = self.classifier(x)

        if self.pointrend:
            
            if self.vgg:
                pointrend_out = self.pointrend_head(x, [features[-4], features[-3]])
            else:
                pointrend_out = self.pointrend_head(x, [features[1]])

        result = F.interpolate(x, x_size[2:], mode='bilinear', align_corners=False)

        if self.training:

            if self.vgg:
                x = self.aux_classifier(features[-2])
            else:
                x = self.aux_classifier(features[0])

            x = F.interpolate(x, x_size[2:], mode='bilinear', align_corners=False)

            if self.pointrend:
                return result, x, pointrend_out
            else:
                return result, x
        else:

            if self.pointrend:
                return pointrend_out
            else:
                return result

def mask_iou(mask1, mask2):
     
    mask1 = mask1.permute(0, 2, 3, 1)
    mask2 = mask2.permute(0, 2, 3, 1)
    mask1 = torch.reshape(mask1 > 0, (-1, mask1.shape[-1])).type(torch.float64)
    mask2 = torch.reshape(mask2 > 0, (-1, mask2.shape[-1])).type(torch.float64)
    area1 = torch.sum(mask1, dim=0)
    area2 = torch.sum(mask2, dim=0)
    intersection = torch.sum(mask1*mask2, dim=0)
    union = area1 + area2 - intersection
    iou = intersection / (union + 1e-6)
    return iou

def compute_miou(model, dl, mean, num_classes, show_progress, ignore_mapped_class=[]):

    ious=[]
    model.learn.model.eval()
    with torch.no_grad():
        for input, target in progress_bar(dl, display=show_progress):
            pred = model.learn.model(input)
            target = target.squeeze(1)
            if ignore_mapped_class != []:
                for k in ignore_mapped_class:
                    pred[:, k] = -1000
                pred = pred.argmax(dim=1)
            else:
                pred = pred.argmax(dim=1)
            mask1 = []
            mask2 = []
            for i in range(pred.shape[0]):
                mask1.append(pred[i].to(model._device) == num_classes[:, None, None].to(model._device))
                mask2.append(target[i].to(model._device) == num_classes[:, None, None].to(model._device))
            mask1 = torch.stack(mask1)
            mask2 = torch.stack(mask2)
            iou = mask_iou(mask1, mask2)
            ious.append(iou.tolist())

    return np.mean(ious, 0)


import traceback
from .._utils.env import raise_fastai_import_error

import_exception = None
try:
    from ._arcgis_model import ArcGISModel, SaveModelCallback, _set_multigpu_callback
    from ._pointcnn_utils import PointCNNSeg, SamplePointsCallback, CrossEntropyPC, accuracy, accuracy_non_zero, AverageMetric
    from .._utils.pointcloud_data import get_device, inference_las, show_results, compute_precision_recall, predict_h5, show_results_tool
    from ._unet_utils import is_no_color
    from fastai.basic_train import Learner
    import torch
    import numpy as np
    from fastai.callbacks import EarlyStoppingCallback
    from functools import partial
    from ._arcgis_model import _EmptyData
    import json
    from pathlib import Path
    from .._utils.common import _get_emd_path
    HAS_FASTAI = True
except Exception as e:
    import_exception = traceback.format_exc()
    HAS_FASTAI = False

class PointCNN(ArcGISModel):

    """
    Model architecture from https://arxiv.org/abs/1801.07791.
    Creates a Point Cloud classification model. 

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            `prepare_data` function.
    ---------------------   -------------------------------------------
    pretrained_path         Optional String. Path where pre-trained model
                            is saved. 
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    encoder_params          Optional dictionary. The keys of the dictionary are 
                            `out_channels`, `P`, `K`, `D` and `m`.

                              Examples:
                                {'out_channels':[16, 32, 64, 96],
                                'P':[-1, 768, 384, 128],
                                'K':[12, 16, 16, 16],
                                'D':[1, 1, 2, 2],
                                'm':8
                                }  

                            Length of `out_channels`, `P`, `K`, `D` should be same.
                            The length denotes the number of layers in encoder.
                              Parameter Explanation
                                - 'out_channels': Number of channels produced by each layer,
                                - 'P': Number of points in each layer,
                                - 'K': Number of K-nearest neighbor in each layer,
                                - 'D': Dilation in each layer,
                                - 'm': Multiplier which is multiplied by each element of out_channel.
    ---------------------   -------------------------------------------
    dropout                 Optional float. This parameter will control overfitting.                          
                            The range of this parameter is [0,1).
    ---------------------   -------------------------------------------
    sample_point_num        Optional integer. The number of points that the model
                            will actually process.     
    =====================   ===========================================

    :returns: `PointCNN` Object
    """

    def __init__(self, data, pretrained_path=None, *args, **kwargs):
        super().__init__(data, None)

        if not HAS_FASTAI:
            raise_fastai_import_error(import_exception=import_exception, message="This model requires module 'torch_geometric' to be installed.", installation_steps=' ')
        
        self._backbone = None
        self.sample_point_num = kwargs.get('sample_point_num', data.max_point)
        self.learn = Learner(data,
                PointCNNSeg(self.sample_point_num, data.c, data.extra_dim, kwargs.get('encoder_params', None), kwargs.get('dropout', None)),
                loss_func=CrossEntropyPC(data.c),
                metrics=[AverageMetric(accuracy)],
                callback_fns=[partial(SamplePointsCallback, sample_point_num=self.sample_point_num)])
        self.encoder_params = self.learn.model.encoder_params

        self.learn.model = self.learn.model.to(self._device)

        if pretrained_path is not None:
            self.load(pretrained_path)

    @classmethod
    def from_model(cls, emd_path, data=None):
        
        """
        Creates an PointCNN model object from a Deep Learning Package(DLPK)
        or Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :returns: `PointCNN` Object
        """      

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd['ModelFile'])
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file
        model_params = emd['ModelParameters']

        try:
            class_mapping = {i['Value'] : i['Name'] for i in emd['Classes']}
            color_mapping = {i['Value'] : i['Color'] for i in emd['Classes']}
        except KeyError:
            class_mapping = {i['ClassValue'] : i['ClassName'] for i in emd['Classes']} 
            color_mapping = {i['ClassValue'] : i['Color'] for i in emd['Classes']}                

        if data is None:
            data = _EmptyData(path=emd_path.parent.parent, loss_func=None, c=len(class_mapping), chip_size=emd['ImageHeight'])
            data.emd_path = emd_path
            data.emd = emd
            for key, value in emd['DataAttributes'].items():
                setattr(data, key, value)
            ## backward compatibility.
            if not hasattr(data, 'class_mapping'):
                data.class_mapping = class_mapping
            if not hasattr(data, 'color_mapping'):
                data.color_mapping = color_mapping
            if not hasattr(data, 'classes'):
                data.classes = list(class_mapping.values())

            data.class_mapping = {int(k): int(v) for k, v in data.class_mapping.items()}
            data.color_mapping = {int(k): v for k, v in data.color_mapping.items()}


            ## Below are the lines to make save function work
            data.chip_size = None
            data._image_space_used = None
            data.dataset_type = 'PointCloud'                 

        return cls(data, **model_params, pretrained_path=str(model_file))

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    def fit(self, epochs=10, lr=None, one_cycle=True, early_stopping=False, checkpoint=True, tensorboard=False, **kwargs):

        """
        Train the model for the specified number of epochs and using the
        specified learning rates
        
        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        epochs                  Required integer. Number of cycles of training
                                on the data. Increase it if underfitting.
        ---------------------   -------------------------------------------
        lr                      Optional float. Learning rate to be used
                                for training the model. If ``lr=None``, an
                                optimal learning rate is automatically deduced
                                for training the model.
        ---------------------   -------------------------------------------
        one_cycle               Optional boolean. Parameter to select 1cycle
                                learning rate schedule. If set to `False` no 
                                learning rate schedule is used.       
        ---------------------   -------------------------------------------
        early_stopping          Optional boolean. Parameter to add early stopping.
                                If set to 'True' training will stop if validation
                                loss stops improving for 5 epochs.        
        ---------------------   -------------------------------------------
        checkpoint              Optional boolean. Parameter to save the best model
                                during training. If set to `True` the best model 
                                based on validation loss will be saved during 
                                training.
        ---------------------   -------------------------------------------
        tensorboard             Optional boolean. Parameter to write the training log.
                                If set to 'True' the log will be saved at
                                <dataset-path>/training_log which can be visualized in
                                tensorboard. Required tensorboardx version=2.1

                                The default value is 'False'.
                                **Note - Not applicable for Text Models
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        iters_per_epoch         Optional integer. The number of iterations 
                                to run during the training phase.
        =====================   ===========================================
                
        """
        iterations = kwargs.get('iters_per_epoch', None)
        from ._pointcnn_utils import IterationStop
        callbacks = kwargs['callbacks'] if 'callbacks' in kwargs.keys() else []
        if iterations is not None:
            del kwargs['iters_per_epoch']
            stop_iteration_cb = IterationStop(self.learn, iterations)
            callbacks.append(stop_iteration_cb)
            kwargs['callbacks'] = callbacks
        self._check_requisites()

        if lr is None:
            print('Finding optimum learning rate.')
            lr = self.lr_find(allow_plot=False)

        if isinstance(lr, slice):
            lr = lr.stop
        
        super().fit(epochs, lr, one_cycle, early_stopping, checkpoint, tensorboard, **kwargs)
        
    @property
    def _model_metrics(self):
        return {'accuracy': self._get_model_metrics()}            

    def _get_model_metrics(self, **kwargs):
        checkpoint = kwargs.get('checkpoint', True)
        if not hasattr(self.learn, 'recorder'):
            return 0.0

        model_accuracy = self.learn.recorder.metrics[-1][0]
        if checkpoint:
            val_losses = self.learn.recorder.val_losses
            model_accuracy = self.learn.recorder.metrics[val_losses.index(min(val_losses))][0]

        return float(model_accuracy)

    def unfreeze(self):
        """
        Unfreezes the earlier layers of the model for
        fine-tuning. Not implemented for PointCNN as
        none of the layers are frozen by default.
        """
        self.learn.unfreeze()        

    def _get_emd_params(self, save_inference_file):
        import random
        _emd_template = {"DataAttributes" : {}, "ModelParameters" : {}}
        _emd_template["Framework"] = "N/A"
        _emd_template["ModelConfiguration"] = "N/A"
        _emd_template["ExtractBands"] = "N/A"
        _emd_template["ModelParameters"]["encoder_params"] = self.encoder_params
        _emd_template["ModelParameters"]["sample_point_num"] = self.sample_point_num

        _emd_template['DataAttributes']['block_size'] = self._data.block_size
        _emd_template['DataAttributes']['max_point'] = self._data.max_point
        _emd_template['DataAttributes']['extra_features'] = self._data.extra_features
        _emd_template['DataAttributes']['extra_dim'] = self._data.extra_dim
        _emd_template['DataAttributes']['class_mapping'] = self._data.class_mapping
        _emd_template['DataAttributes']['color_mapping'] = self._data.color_mapping
        _emd_template['DataAttributes']['remap'] = self._data.remap
        _emd_template['DataAttributes']['classes'] = self._data.classes

        _emd_template['Classes'] = []
        class_data = {}
        for i, class_name in enumerate(self._data.classes):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[i]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)] if is_no_color(self._data.color_mapping) else \
                self._data.color_mapping[class_name]
            class_data["Color"] = np.array(color).astype(int).tolist()
            _emd_template['Classes'].append(class_data.copy())

        return _emd_template
        
    def show_results(self, rows=2, **kwargs):

        """
        Displays the results from your model on the validation set
        with ground truth on the left and predictions on the right.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional rows. Number of rows to show. Default
                                value is 2 and maximum value is the `batch_size`
                                passed in `prepare_data`. 
        =====================   ===========================================

        **kwargs**

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        color_mapping           Optional dictionary. Mapping from class value
                                to RGB values. Default value example:
                                {0:[220,220,220], 2:[255,0,0], 6:[0,255,0]}.
        ---------------------   -------------------------------------------
        mask_class              Optional list of integers. Array containing
                                class values to mask. Use this parameter to 
                                display the classes of interest.
                                Default value is []. 
                                Example: All the classes are in [0, 1, 2]
                                to display only class `0` set the mask class
                                parameter to be [1, 2]. List of all classes
                                can be accessed from `data.classes` attribute
                                where `data` is the `Databunch` object returned
                                by `prepare_data` function.    
        ---------------------   -------------------------------------------
        width                   Optional integer. Width of the plot. Default 
                                value is 750.
        ---------------------   -------------------------------------------
        height                  Optional integer. Height of the plot. Default
                                value is 512.   
        ---------------------   -------------------------------------------
        max_display_point       Optional integer. Maximum number of points
                                to display. Default is 20000. A warning will
                                be raised if the total points to display exceeds
                                this parameter. Setting this parameter will
                                randomly sample the specified number of points
                                and once set, it will be used for future uses.
        =====================   ===========================================
        """

        if self._data.pc_type == 'PointCloud_TF':
            return show_results(self, rows, **kwargs)
        elif self._data.pc_type == 'PointCloud':
            return show_results_tool(self, rows, **kwargs)

    def predict_las(self, path, output_path=None, print_metrics=False, **kwargs):

        """
        Predicts and writes the resulting las file on the disk.
        The block size which was used for training will be used for prediction.
        Coordinate system for the inferencing data & trained model's training
        data should be the same.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        path                    Required string. The path to folder where the las
                                files which needs to be predicted are present.   
        ---------------------   -------------------------------------------
        output_path             Optional string. The path to folder where to dump
                                the resulting las files. Defaults to `results` folder
                                in input path.  
        ---------------------   -------------------------------------------
        print_metrics           Optional boolean. If True, precision, recall and
                                f1_score are also calculated and reported.
                                Defaults to False.
        =====================   ===========================================                                

        **kwargs**

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        remap_classes           Optional dictionary {int:int}. Mapping from  
                                class values to user defined values. Please query 
                                `pointcnn._data.classes` to get the class values
                                on which the model is trained on.
                                Default is {}.
        ---------------------   -------------------------------------------
        selective_classify      Optional list of integers. If passed, predict_las 
                                will selectively classify only those points 
                                belonging to the specified class-codes. Other 
                                points in the input point clouds will retain 
                                their class-codes. 
                                Please query `pointcnn._data.classes` to get 
                                the class values on which the model is trained 
                                on. If `remap_classes` is specified, the new 
                                mapped values will be used for classification. 
                                Default value is [].
        ---------------------   -------------------------------------------
        preserve_classes        Optional list of integers. A list of classes
                                from the input data, that should be preserved 
                                in the predicted output.
                                If a point in the input data belongs to any 
                                of the classes mentioned in this list, its 
                                class-code won't be updated with the model's 
                                predicted class. 
                                Example: If preserve_classes=[2,6]. The 
                                class-code of a point won't be updated with 
                                the predicted class, if it's 2 or 6. 
                                Default: [].
        =====================   ===========================================
        
        :returns: Path where files are dumped.
        """
        
        return inference_las(path, self, output_path, print_metrics, **kwargs)

    def compute_precision_recall(self):
        
        """
        Computes precision, recall and f1-score on the validation sets.
        """

        return compute_precision_recall(self)

    def predict_h5(self, path, output_path=None, **kwargs):
        """
        Predicts and writes the resulting las file on the disk. 
        The block size which was used for training will be used for prediction.
        Coordinate system for the inferencing data & trained model's training
        data should be the same.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        path                    Required string. The path to folder where the h5
                                files which needs to be predicted are present.   
        ---------------------   -------------------------------------------
        output_path             Optional string. The path to folder where to dump
                                the resulting h5 block files. Defaults to `results`
                                folder in input path.
        =====================   ===========================================
        
        :returns: Path where files are dumped.
        """
        
        return predict_h5(self, path, output_path)        
        



import os
os.environ["USE_TF"] = "0"
os.environ["USE_TORCH"] = "1"
import re
import traceback
import collections
from functools import partial

HAS_TRANSFORMER = True

try:
    import torch
    import pandas as pd
    from fastai.data_block import DataBunch
    from fastai.basic_data import DatasetType
    from torch.utils.data import Dataset, DataLoader
    from transformers import PreTrainedTokenizer, AutoTokenizer
    from fastai.text import List, BaseTokenizer, Vocab, Collection, SortishSampler, SortSampler
    from fastai.torch_core import to_data, BatchSamples, Tuple, LongTensor
except Exception as e:
    transformer_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_TRANSFORMER = False

    class BaseTokenizer:
        pass

    class Vocab:
        pass


def process_text(text):
    text = re.sub(' +', ' ', text)
    text = re.sub(' ([@.#$/:-]) ?', r'\1', text)
    # Reomve leading and trailing special characters
    while text and text[0] in ['.', ',', '-', ':', '@', '$']: text = text[1:]
    while text and text[-1] in ['.', ',', '-', ':', '@', '$']: text = text[:-1]
    return text.strip()


models_subwords_map = {
    "distilbert": ("##", False),
    "bert": ("##", False),
    "electra": ("##", False),
    "xlnet": ("▁", True),
    "camembert": ("▁", True),
    "albert": ("▁", True),
    "xlm-roberta": ("▁", True),
    "roberta": ("Ġ", True),
    "longformer": ("Ġ", True),
    "bart": ("Ġ", True),
    "mobilebert": ("##", False)
}


class TransformersBaseTokenizer(BaseTokenizer):
    """
    Wrapper around PreTrainedTokenizer to be compatible with fast.ai
    """
    def __init__(self, pretrained_tokenizer: PreTrainedTokenizer, lang="en", seq_len=512, **kwargs):
        super().__init__(lang)
        self._pretrained_tokenizer = pretrained_tokenizer
        self._seq_len = seq_len
        self.max_seq_len = min(pretrained_tokenizer.model_max_length, self._seq_len)

    def __call__(self, *args, **kwargs):
        return self

    # new improved tokenizer which can support any transformer model
    def tokenizer(self, t:str) -> List[str]:
        ids = self._pretrained_tokenizer.encode(t, max_length=self.max_seq_len, truncation=True)
        tokens = self._pretrained_tokenizer.convert_ids_to_tokens(ids)
        return tokens


class TransformersVocab(Vocab):
    """
    Contain the correspondence between numbers and tokens and numericalize.
    """
    def __init__(self, tokenizer: PreTrainedTokenizer):
        super().__init__(itos=[])
        self.tokenizer = tokenizer
        self.special_token_map = self.tokenizer.special_tokens_map
        self.has_cls_token = True if self.special_token_map.get('cls_token') else False
        self.has_sep_token = True if self.special_token_map.get('sep_token') else False
        self.has_unk_token = True if self.special_token_map.get('unk_token') else False
        self.has_pad_token = True if self.special_token_map.get('pad_token') else False

    def numericalize(self, t: Collection[str]) -> List[int]:
        """
        Convert a list of tokens `t` to their ids.
        """
        return self.tokenizer.convert_tokens_to_ids(t)

    def textify(self, nums: Collection[int], sep=' ') -> str:
        """
        Convert a list of `nums` to their tokens.
        """
        # text = self.tokenizer.decode(nums, skip_special_tokens=True)
        text = self.tokenizer.decode(nums)
        if self.has_cls_token: text = text.replace(self.tokenizer.cls_token, '')
        if self.has_sep_token: text = text.replace(self.tokenizer.sep_token, '')
        if self.has_unk_token: text = text.replace(self.tokenizer.unk_token, '')
        if self.has_pad_token: text = text.replace(self.tokenizer.pad_token, '')
        text = re.sub(' +', ' ', text)
        return text.strip()

    def __getstate__(self):
        return {'itos': self.itos, 'tokenizer': self.tokenizer}

    def __setstate__(self, state: dict):
        self.itos = state['itos']
        self.tokenizer = state['tokenizer']
        self.stoi = collections.defaultdict(int, {v: k for k, v in enumerate(self.itos)})


class TransformerNERDataset(Dataset):
    def __init__(self, tokens_collection, tags_collection, tokenizer, label2id, model_type="bert", seq_length=512):
        self.tokens_collection = tokens_collection
        self.tags_collection = tags_collection
        self.tokenizer = tokenizer
        self.label2id = label2id
        self.id2label = {y: x for x, y in self.label2id.items()}
        self.model_type = model_type
        self.ignore_token_list = ["\n"]
        self.otag_id = self.label2id['O'] if 'O' in self.label2id else self.label2id['o']
        self.max_seq_len = min(tokenizer.model_max_length, seq_length)

    def __len__(self):
        return len(self.tokens_collection)

    def __getitem__(self, index):
        tokens = self.tokens_collection[index]
        tags = self.tags_collection[index]
        ids, target_tags = [], []

        tokens = [x for x in tokens if x not in self.ignore_token_list]
        if self.model_type in ["roberta", "bart", "longformer"]:
            # tokens = [f" {x}" if index != 0 else x for index, x in enumerate(tokens)]
            tokens = [f" {x}" for index, x in enumerate(tokens)]

        for index, token in enumerate(tokens):
            # Transformer tokenizer can split the input word at sub-word level.
            # For example - `uninvited` can be broken into: 'un', '##in', '##vi', '##ted'
            token_breakup_list = self.tokenizer.encode(token, add_special_tokens=False)
            num_items = len(token_breakup_list)
            ids.extend(token_breakup_list)
            tag_id = self.label2id[tags[index]]
            target_tags.extend([tag_id] * num_items)

        # self.max_seq_len - 2, because we will be adding a [CLS] & [SEP] token when creating the DataBunch
        ids = ids[:self.max_seq_len - 2]
        target_tags = target_tags[:self.max_seq_len - 2]

        if self.model_type in ["xlnet"]:
            ids = ids + [self.tokenizer.sep_token_id, self.tokenizer.cls_token_id]
            target_tags = target_tags + [self.otag_id, self.otag_id]
        elif self.model_type in ["xlm", "flaubert"]:
            ids = [self.tokenizer.bos_token_id] + ids + [self.tokenizer.sep_token_id]
            target_tags = [self.otag_id] + target_tags + [self.otag_id]
        else:
            ids = [self.tokenizer.cls_token_id] + ids + [self.tokenizer.sep_token_id]
            target_tags = [self.otag_id] + target_tags + [self.otag_id]

        return {'ids': ids, 'target': target_tags, 'otag_id': self.otag_id}


def collate_func(samples: BatchSamples, tokenizer: PreTrainedTokenizer, seq_len: int=512):
    # Function that collect samples and adds padding
    samples = to_data(samples)
    batch_sentences = [x['ids'] for x in samples]
    batch_target = [x['target'] for x in samples]
    otag_id = samples[0]['otag_id']

    # [CLS] & [SEP] token are already added from the items returned from TransformerNERDataset
    max_len_batch = max([len(x) for x in batch_target])
    max_seq_len = min(tokenizer.model_max_length, seq_len)
    max_len = min(max_len_batch, max_seq_len)

    res_x = tokenizer.batch_encode_plus(batch_sentences, max_length=max_len, padding=True, truncation=True,
                                        is_split_into_words=True, add_special_tokens=False, return_tensors="pt")

    input_ids, attention_masks = res_x.get("input_ids"), res_x.get("attention_mask")
    token_type_ids = res_x.get("token_type_ids")

    res_y = torch.zeros(len(samples), max_len).long() + otag_id

    pad_first = True if tokenizer.padding_side == 'left' else False
    for i, s in enumerate(batch_target):
        if pad_first:
            res_y[i, -len(s):] = LongTensor(s)
        else:
            res_y[i, :len(s):] = LongTensor(s)

    # res_y is added in the list as the model has the ability to return the loss if the labels are provided
    if token_type_ids is not None:
        return [input_ids, attention_masks, token_type_ids, res_y], [res_y, attention_masks]
    else:
        return [input_ids, attention_masks, res_y], [res_y, attention_masks]


class TransformerNERDataBunch(DataBunch):
    "Create a `TextDataBunch` suitable for training an RNN classifier."

    @classmethod
    def create(cls, train_ds, valid_ds, tokenizer: PreTrainedTokenizer, bs: int = 4, val_bs: int = None,
               device: torch.device = None, no_check: bool = False, seq_len: int = 512, **dl_kwargs) -> DataBunch:
        "Function that transform the `datasets` in a `DataBunch`. Passes `**dl_kwargs` on to `DataLoader()`"
        if val_bs is None: val_bs = bs
        # write own collate function
        collate_fn = partial(collate_func, tokenizer=tokenizer, seq_len=seq_len)
        train_sampler = SortishSampler(train_ds, key=lambda t: len(train_ds[t]['target']), bs=bs)
        train_dl = DataLoader(train_ds, batch_size=bs, sampler=train_sampler, drop_last=False, **dl_kwargs)
        dataloaders = [train_dl]

        sampler = SortSampler(valid_ds, key=lambda t: len(valid_ds[t]['target']))
        dataloaders.append(DataLoader(valid_ds, batch_size=val_bs, sampler=sampler, drop_last=False, **dl_kwargs))
        return cls(*dataloaders, device=device, collate_fn=collate_fn, no_check=no_check)

    @property
    def is_empty(self) -> bool:
        return not ((self.train_dl and len(self.train_ds) != 0) or
                    (self.valid_dl and len(self.valid_ds) != 0) or
                    (self.test_dl and len(self.test_ds) != 0))

    def show_batch(self, rows: int = 5, ds_type:DatasetType=DatasetType.Train, reverse:bool=False, **kwargs):
        x, y = self.one_batch(ds_type, True, True)
        batch_tokens = x[0].tolist()
        batch_labels = y[0].tolist()
        num_items = min(rows, self.dl(DatasetType.Train).batch_size)
        results = get_results(batch_tokens, batch_labels, self.train_ds.tokenizer,
                              self.train_ds.id2label, self.train_ds.model_type, num_items)
        df = pd.DataFrame(results, )
        df.fillna("", inplace=True)
        return df


def get_previous_tokens(token_ids, index, model_type, tok):
    res = models_subwords_map.get(model_type)
    if res is None: return [token_ids[index]]

    i, whole_word_ids = index, []
    spl_char, start_token_contain_spl_char = res
    while i > 0:
        curr_token = tok.convert_ids_to_tokens(token_ids[i])
        whole_word_ids.insert(0, token_ids[i])
        if curr_token.startswith(spl_char):
            if start_token_contain_spl_char: break
        else:
            if not start_token_contain_spl_char: break
        i -= 1
    return whole_word_ids


def get_next_tokens(token_ids, index, model_type, tok):
    res = models_subwords_map.get(model_type)
    if res is None: return []

    # i, whole_word_ids = index+1, [token_ids[index]]
    i, whole_word_ids = index + 1, []
    spl_char, start_token_contain_spl_char = res

    while i < len(token_ids):
        curr_token = tok.convert_ids_to_tokens(token_ids[i])
        if curr_token.startswith(spl_char):
            if start_token_contain_spl_char: break
        else:
            if not start_token_contain_spl_char: break
        whole_word_ids.append(token_ids[i])
        i += 1
    return whole_word_ids


def get_results(batch_tokens, batch_labels, tokenizer, id2label, model_type, num_items):
    results = []
    for index, (tokens, labels) in enumerate(zip(batch_tokens, batch_labels)):
        labels = [id2label[x] for x in labels]
        if index >= num_items: break
        prev_label, token_list, entities = labels[0], [(0, tokens[0])], []
        for token_index, (token, label) in enumerate(list(zip(tokens[1:], labels[1:]))):
            label = label.split("-")[-1]
            if label == 'O':
                if prev_label != 'O': entities.append((token_list, prev_label))
                token_list, prev_label = list(), label

            if prev_label == label:
                token_list.append((token_index + 1, token))
            else:
                if prev_label != 'O': entities.append((token_list, prev_label))
                token_list = list()
                prev_label = label
                token_list.append((token_index + 1, token))
        if token_list:
            entities.append((token_list, prev_label))

        # entities is a list [([(4, 1196), (5, 345)], "Address"), ([(34, 444), (35, 3834)], "Person")]
        entity_list = []
        for item in entities:
            token_list, entity = item
            # print("Before - ", tokenizer.decode([x[1] for x in token_list]))
            prev_tokens = get_previous_tokens(tokens, token_list[0][0], model_type, tokenizer)
            end_tokens = get_next_tokens(tokens, token_list[-1][0], model_type, tokenizer)

            token_list = prev_tokens + [x[1] for x in token_list[1:]] + end_tokens
            # print("After - ", tokenizer.decode(token_list))
            entity_text = process_text(tokenizer.decode(token_list, skip_special_tokens=True))
            if len(entity_text) > 2: entity_list.append((entity_text, entity))

        entity_dict = {}
        text = process_text(tokenizer.decode(batch_tokens[index], skip_special_tokens=True))
        entity_dict["Text"] = text
        _ = [entity_dict.setdefault(x[1], []).append(x[0]) for x in entity_list if x[0]]
        results.append(entity_dict)
    return results
